'''

LightCurve objects for TransitFit

'''

import numpy as np
from .detrending_funcs import NthOrderDetrendingFunction
from .detrender import DetrendingFunction
from copy import deepcopy
import csv
from scipy.optimize import curve_fit
import os

class LightCurve:
    '''
    The transit data which we are trying to fit.

    The LightCurve is designed
    to simplify dealing with detrending etc across multiple data sets. It
    allows us to clearly keep values pointed at specific data sets.

    Parameters
    ----------
    times : array_like, shape (X, )
        The times of observation
    flux : array_like, shape (X, )
        The fluxes
    errors : array_like, shape (X, )
        The absolute uncertainty on the fluxes
    telescope_idx : int, optional
        The telescope index associated with this light curve
    filter_idx : int, optional
        The filter index associated with this light curve
    epoch_idx : int, optional
        The epoch index associated with this light curve
    curve_labels : array_like, shape (X, ), optional
        Used to identify if different data points come from different
        observations. Useful when combining curves if you want to undo that!
    telescope_array : array_like, shape (X, ), optional
        Array of the telescope indices for each data point. Useful when dealing
        with combined curves.
    filter_array : array_like, shape (X, ), optional
        Array of the filter indices for each data point. Useful when dealing
        with combined curves.
    epoch_array : array_like, shape (X, ), optional
        Array of the epoch indices for each data point. Useful when dealing
        with combined curves.
    '''
    def __init__(self, times, flux, errors, telescope_idx=None,
                 filter_idx=None, epoch_idx=None, curve_labels=None,
                 telescope_array=None, filter_array=None, epoch_array=None):


        times = np.array(times)
        flux = np.array(flux)
        errors = np.array(errors)

        if not times.ndim == 1:
            raise ValueError('Times, flux, and errors have {} dimensions. They should only have 1!'.format(times.ndims))

        self.times = times
        self.flux = flux
        self.errors = errors

        ###############################################################
        # Identifying indices - for unqiuely identifying a light curve:

        # These are quick-access references for non-combined curves
        self.telescope_idx = telescope_idx
        self.filter_idx = filter_idx
        self.epoch_idx = epoch_idx

        # These will be arrays which allow individual light curves to be
        # separated if this is a combined curve. Will be used in outputs too.
        if telescope_array is None:
            if telescope_idx is None:
                self._telescope_array = np.full(times.shape, None)
            else:
                self._telescope_array = np.ones(times.shape) * telescope_idx
        else:
            telescope_array = np.array(telescope_array)
            if not telescope_array.shape == times.shape:
                raise ValueError('telescope_array has shape {}, but requires shape {}'.format(telescope_array.shape, times.shape))
            self._telescope_array = telescope_array

        if filter_array is None:
            if filter_idx is None:
                self._filter_array = np.full(times.shape, None)
            else:
                self._filter_array = np.ones(times.shape) * filter_idx
        else:
            filter_array = np.array(filter_array)
            if not filter_array.shape == times.shape:
                raise ValueError('filter_array has shape {}, but requires shape {}'.format(filter_array.shape, times.shape))
            self._filter_array = filter_array

        if epoch_array is None:
            if epoch_idx is None:
                self._epoch_array = np.full(times.shape, None)
            else:
                self._epoch_array = np.ones(times.shape) * epoch_idx
        else:
            epoch_array = np.array(epoch_array)
            if not epoch_array.shape == times.shape:
                raise ValueError('epoch_array has shape {}, but requires shape {}'.format(epoch_array.shape, times.shape))
            self._epoch_array = epoch_array

        # These are used to label the data so that combined lightcurves
        # can be quickly untangled.
        if curve_labels is None:
            self.curve_labels = np.zeros(times.shape)
        else:
            curve_labels = np.array(curve_labels)
            if not curve_labels.shape == times.shape:
                raise ValueError('curve_labels has shape {}, but requires shape {}'.format(curve_labels.shape, times.shape))
            self.curve_labels = curve_labels

        self.detrend = False
        self.normalise = False

    def set_detrending(self, method, order=None, function=None,
                       method_idx=None):
        '''
        Sets detrending method

        Parameters
        ----------
        method : str
            Accepted
                - 'nth order'
                - 'custom'
        order : int, optional
            The order of detrending to use if method is 'nth order'. Must be
            provided.
        function : function, optional
            If method is 'custom', this is the custom detrending function to
            use.
        '''
        if method.lower() in ['nth order', 'nthorder', 'nth_order']:
            if order is None:
                raise ValueError('You need to provide a detrending order!')

            function = NthOrderDetrendingFunction(order)
            self.detrending_function = DetrendingFunction(function)
            self.n_detrending_params = order
        elif method.lower() == 'custom':
            if function is None:
                raise ValueError('You must provide the custom detrending function')
            self.detrending_function = DetrendingFunction(function,method.lower())
            self.n_detrending_params = self.detrending_function.n_required_args - 1
        elif method.lower() == 'off':
            self.detrending_method_idx = method_idx
            self.detrend = False
            return
        else:
            raise ValueError('Unrecognised method {}'.format(method))

        self.detrending_method_idx = method_idx
        self.detrend = True

    def set_normalisation(self,normalise_limits):
        '''
        Turns on normalisation. Also estimates limits and a best initial guess.

        Parameters
        ----------
        default_low : float, optional
            The lowest value to consider as a multiplicative normalisation
            constant. Default is 0.1.

        Returns
        -------
        median_factor : float
            The initial best guess factor
        low_factor : float
            The low limit on the normalisation factor
        high_factor : float
            The high limit on the normalisation factor

        Notes
        -----
        We use
            ``0.5/f_max <= c_n <= 1.5/f_min``
        as the default range, where f_median is the median flux value.
        '''
        self.normalise = True

        if normalise_limits is None:
                return self.estimate_normalisation_limits()
        else:
                return np.average(normalise_limits), normalise_limits[0], normalise_limits[1]
        
    def set_error_scaling(self,scaling_limits):
        '''
        Turns on normalisation. Also estimates limits and a best initial guess.

        Parameters
        ----------
        default_low : float, optional
            The lowest value to consider as a multiplicative normalisation
            constant. Default is 0.1.

        Returns
        -------
        median_factor : float
            The initial best guess factor
        low_factor : float
            The low limit on the normalisation factor
        high_factor : float
            The high limit on the normalisation factor

        Notes
        -----
        We use
            ``0.5/f_max <= c_n <= 1.5/f_min``
        as the default range, where f_median is the median flux value.
        '''
        self.error_scaling = True

        return np.average(scaling_limits), scaling_limits[0], scaling_limits[1]

    def estimate_normalisation_limits(self):
        '''
        Estimates the range for fitting a normalisation constant, and also
        finds a reasonable first guess.

        Returns
        -------
        median_factor : float
            The initial best guess factor
        low_factor : float
            The low limit on the normalisation factor
        high_factor : float
            The high limit on the normalisation factor

        Notes
        -----
        We use
            ``0.5/f_max <= c_n <= 1.5/f_min``
        as the default range.
        '''
        median_factor = 1/np.median(self.flux)

        # updated v0.10.0 - gives a narrower prior
        low_factor = 0.5/np.max(self.flux)
        high_factor = 1.5/np.min(self.flux)

        return median_factor, low_factor, high_factor

    """def detrend_flux(self, d, norm=1, force_normalise=False):
        '''
        When given a normalisation constant and some detrending parameters,
        will return the detrended flux and errors

        Parameters
        ----------
        d : array_like, shape(n_detrending_params,)
            Array of the detrending parameters to use
        norm : float, optional
            The normalisation constant to use. Default is 1
        force_normalise : bool, optional
            Override to allow normalisation to be forced if the LightCurve does
            not have normalisation initialised. Default is False.

        Returns
        -------
        detrend_flux : array_like, shape(num_times)
            The detrended flux
        detrended_errors : array_like, shape(num_times)
            The errors on the detrended flux
        '''

        if self.detrend and d is not None:
            detrended_flux = self.detrending_function(self, *d)
        else:
            detrended_flux = deepcopy(self.flux)
        detrended_errors = deepcopy(self.errors)

        if self.normalise or force_normalise:
            detrended_flux *= norm
            detrended_errors *= norm

        return detrended_flux, detrended_errors"""

    def detrend_flux(self, d_new, norm=1, force_normalise=False):
        '''
        When given a normalisation constant and some detrending parameters,
        will return the detrended flux and errors

        Parameters
        ----------
        d : array_like, shape(n_detrending_params,)
            Array of the detrending parameters to use
        norm : float, optional
            The normalisation constant to use. Default is 1
        force_normalise : bool, optional
            Override to allow normalisation to be forced if the LightCurve does
            not have normalisation initialised. Default is False.

        Returns
        -------
        detrend_flux : array_like, shape(num_times)
            The detrended flux
        detrended_errors : array_like, shape(num_times)
            The errors on the detrended flux
        '''
        #self.t0 = t0
        d=d_new[:-2]
        if self.detrend and d is not None:
            detrended_flux = self.detrending_function(self, *d_new)
        else:
            detrended_flux = deepcopy(self.flux)
        detrended_errors = deepcopy(self.errors)

        if self.normalise or force_normalise:
            detrended_flux *= norm
            detrended_errors *= norm

        return detrended_flux, detrended_errors

    """def create_detrended_LightCurve(self, d, norm):
        '''
        Creates a detrended LightCurve using detrend_flux() and returns it
        '''
        detrended_flux, detrended_errors = self.detrend_flux(d, norm)

        return LightCurve(self.times, detrended_flux, detrended_errors,
                          self.telescope_idx, self.filter_idx,
                          self.epoch_idx, self.curve_labels,
                          self._telescope_array, self._filter_array,
                          self._epoch_array)"""

    def create_detrended_LightCurve(self, d_new, norm, escale=None):
        '''
        Creates a detrended LightCurve using detrend_flux() and returns it
        '''
        detrended_flux, detrended_errors = self.detrend_flux(d_new, norm)
        
        if escale is not None:
            
            detrended_errors = np.sqrt(np.power(detrended_errors,2) + np.power(escale*detrended_flux,2))

        return LightCurve(self.times, detrended_flux, detrended_errors,
                          self.telescope_idx, self.filter_idx,
                          self.epoch_idx, self.curve_labels,
                          self._telescope_array, self._filter_array,
                          self._epoch_array)

    def fold(self, t0, period, base_t0=None):
        '''
        Folds the LightCurve so that all times are between t0 - period/2 and
        t0 + period/2.

        If base_t0 is provided, this will ensure that the folded lightcurve is
        centred on base_t0. This is to allow for ttv mode where the differences
        between the retrieved t0 for each epoch must be accounted for.

        returns a new LightCurve
        '''
        if base_t0 is None:
            base_t0 = t0

        #ttv_term = t0 - ((t0 - (base_t0+period/2))//period) * period - base_t0
        #ttv_term = self.times - ((self.times - (t0 + period/2))//period) * period - base_t0

        #times = self.times - ((self.times - (t0 + period/2))//period) * period - ttv_term
        phase = self.get_phases(t0, period)

        n = (self.times - (t0 - 0.5*period))//period

        times = base_t0 + period * (phase + n)# - 0.5)

        return LightCurve(times, self.flux, self.errors, self.telescope_idx,
                          self.filter_idx, self.epoch_idx, self.curve_labels,
                          self._telescope_array, self._filter_array,
                          self._epoch_array)

    def combine(self, lightcurves, telescope_idx=None, filter_idx=None,
                epoch_idx=None):
        '''
        Combines the LightCurve with a list of other lightcurves which are passed to it
        and returns a new lightcurve containing all the data of the input
        curves. Note that you should probably only do this with lightcurves
        which have already been detrended, otherwise you might struggle
        to detrend the combined one.
        '''
        times = self.times
        flux = self.flux
        errors = self.errors

        labels = self.curve_labels
        telescope_labels = self._telescope_array
        filter_labels = self._filter_array
        epoch_labels = self._epoch_array

        starting_label = labels.max() + 1
        for i, lightcurve in enumerate(lightcurves):
            times = np.hstack((times, lightcurve.times))
            flux = np.hstack((flux, lightcurve.flux))
            errors = np.hstack((errors, lightcurve.errors))
            labels = np.hstack((labels, np.ones(lightcurve.times.shape) * (starting_label + i)))
            telescope_labels = np.hstack((telescope_labels, lightcurve._telescope_array))
            filter_labels = np.hstack((filter_labels, lightcurve._filter_array))
            epoch_labels = np.hstack((epoch_labels, lightcurve._epoch_array))

        return LightCurve(times, flux, errors, telescope_idx, filter_idx,
                          epoch_idx, labels, telescope_labels, filter_labels,
                          epoch_labels)

    def decombine(self):
        '''
        Splits a lightcurve according to its curve labels.

        Inverse of combine
        '''
        lightcurves = []

        for label in np.unique(self.curve_labels):
            mask = self.curve_labels == label
            times = self.times[mask]
            flux = self.flux[mask]
            errors = self.errors[mask]
            telescope_array = self._telescope_array[mask]
            filter_array = self._filter_array[mask]
            epoch_array = self._epoch_array[mask]

            lightcurves.append(LightCurve(times, flux, errors,
                                          telescope_array[0], filter_array[0],
                                          epoch_array[0], None, telescope_array,
                                          filter_array, epoch_array))

        return lightcurves

    def split(self, t0, P, t14, cutoff=0.25, window=None):
        '''
        Splits the LightCurve into multiple LightCurves containing a single
        transit. This is useful for dealing with multi-epoch observations which
        contain TTVs, or have long term periodic trends, since single-epoch
        observation trends can be approximated with polynomial fits.

        Parameters
        ----------
        t0 : float
            The centre of a transit
        P : float
            The estimated period of the planet in days
        t14 : float, optional
            The approximate transit duration in minutes.
        cutoff : float, optional
            If there are no data within t14 * cutoff of t0, a period will be
            discarded. Default is 0.25
        window : float, optional
            If provided, data outside of the range [t0 ± (0.5 * t14) * window]
            will be discarded.

        Returns
        -------
        lightcurves : list
            A list of LightCurve objects

        Notes
        -----
        Will reset the telescope, filter and epoch indices to None.
        '''
        # Work out how many periods are contained in the current LightCurve
        n_periods = np.ceil((self.times.max() - self.times.min())/P)
        n_periods = int(n_periods) +1

        # Make sure that t0 will fall in the first new epoch
        t0 = t0 - ((t0 - self.times.min())//P * P)

        # Convert t14 to days:
        t14 /= 60 * 24

        # Work out the times, flux, and errors for each epoch
        t_new = [[] for i in range(n_periods)]
        f_new = [[] for i in range(n_periods)]
        err_new = [[] for i in range(n_periods)]

        # Loop through each data point and assign to the correct epoch
        for i in range(len(self.times)):
            t = self.times[i]
            f = self.flux[i]
            err = self.errors[i]

            for j in range(n_periods):
                if t < t0 + (2 * j + 1) * P/2 and t > t0 - P/2:
                    t_new[j].append(t)
                    f_new[j].append(f)
                    err_new[j].append(err)
                    break

        # Convert to np arrays
        for i in range(len(t_new)):
            t_new[i] = np.array(t_new[i])
            f_new[i] = np.array(f_new[i])
            err_new[i] = np.array(err_new[i])

        # Now we have all the data for the epochs, make the new LightCurves
        lightcurves = []
        for i in range(n_periods):
            if not len(t_new[i]) == 0:
                # Make sure we don't produce empty curves

                # Now a bunch more checks. Need to ensure there is data around
                # the expected t0 (ie we have a transit!). We keep the transit
                # if there is data within t14*cutoff of the predicted t0
                # (found by folding each epoch)
                if np.any(abs((t_new[i]- ((t_new[i] - t0)//P) * P) - t0) <= t14*cutoff):

                    # Now we cut out the window:
                    if window is None:
                        mask = np.ones(len(t_new[i]), bool)
                    else:
                        mask = abs((t_new[i] - ((t_new[i] - (t0 - P/2))//P) * P) - t0) <= (window * t14 * 0.5)

                    new_curve = LightCurve(t_new[i][mask], f_new[i][mask], err_new[i][mask])

                    lightcurves.append(new_curve)
                else:
                    print('Light curve {} discarded'.format(i))

        return lightcurves

    def save(self, filepath):
        '''
        Saves the LightCurve to a .csv file
        '''
        if not filepath[-4:] == '.csv':
            filepath += '.csv'

        os.makedirs(os.path.dirname(filepath), exist_ok=True)

        write_dict = []
        for j, tj in enumerate(self.times):
            write_dict.append({'Time' : tj,
                               'Flux' : self.flux[j],
                               'Flux_err' : self.errors[j]})

        with open(filepath, 'w') as f:
            columns = ['Time', 'Flux', 'Flux_err']
            writer = csv.DictWriter(f, columns)
            writer.writeheader()
            writer.writerows(write_dict)

    def get_phases(self, t0, P):
        '''
        Converts times into phase given t0 and P values, with t0 at phase=0
        '''
        n = (self.times - (t0 - 0.5*P))//P


        return (self.times-t0)/P - n# + 0.5


    def bin(self, cadence, residuals=None):
        '''
        Bins the light curve to a given observation cadence

        Parameters
        ----------
        cadence : float
            The observation cadence to bin to in the units of self.times
        residuals : array_like, shape (n_times,)
            If provided, will also bin any model residuals

        Returns
        -------
        time : array_like
            The time centers of each bin
        flux : array_like
            The flux in each bin, calculated as the weighted mean of all flux
            values in the bin
        err : array_like
            The error on the flux of each bin
        '''
        if residuals is not None:
            residuals = np.array(residuals)

        # Calculate the number of bins we need, and the size of each bin
        obs_length = self.times.max() - self.times.min()

        n_bins = int((obs_length)/cadence)
        bin_size = obs_length / n_bins

        times = np.full(n_bins, None)
        flux = np.full(n_bins, None)
        err = np.full(n_bins, None)
        binned_residuals = np.full(n_bins, None)

        points_in_bin = np.zeros(n_bins)

        # Now we can bin the data
        for i in range(n_bins):
            bin_upper = (i+1) * bin_size + self.times.min()
            bin_lower = i * bin_size + self.times.min()
            bin_time = (bin_upper + bin_lower)/2

            mask = (bin_lower <= self.times) * (self.times < bin_upper)

            bin_flux = self.flux[mask]

            bin_err = self.errors[mask]

            if len(bin_flux > 0):
                if residuals is not None:
                    bin_residuals = residuals[mask]
                    binned_residuals[i] = np.average(bin_residuals, weights=1/(bin_err**2))

                times[i] = bin_time
                flux[i] = np.average(bin_flux, weights=1/(bin_err**2))
                err[i] = 1/np.sqrt(np.sum(1/(bin_err**2)))

            points_in_bin[i] = np.sum(mask)

        return_mask = (points_in_bin > 0)

        if residuals is None:
            return times[return_mask], flux[return_mask], err[return_mask]
        else:
            return times[return_mask], flux[return_mask], err[return_mask], binned_residuals[return_mask]

    def __eq__(self, other):
        '''
        Checks two LightCurves are the same. We are only checking the times,
        fluxes and errors.
        '''
        if not isinstance(other, LightCurve):
            return False

        if not self.times.shape == other.times.shape:
            return False
        if not np.all(self.times == other.times):
            return False

        if not self.flux.shape == other.flux.shape:
            return False
        if not np.all(self.flux == other.flux):
            return False

        if not self.errors.shape == other.errors.shape:
            return False
        if not np.all(self.errors == other.errors):
            return False

        return True



"""
This python file takes output_parameters folder as input and provides 
asymmetric errors using the distribution of samples. 
"""
import glob
import pickle
import numpy as np
import pandas as pd
from statsmodels.stats import weightstats


def get_quantiles_on_best_val(samples, weights, best_val):
    """Generates lower and upper limit of errors for the best values.
    Sorts the samples and corresponding weights. 
    Gets value of samples such that they encompass 68.27% of the samples
    by weight on both sides of the best value.

    Args:
        samples (array): the sampled values for the parameter from dynesty
        weights (array): weights of the samples
        best_val (float): best value among the samples

    Returns:
        tuple: the lower and upper error on the best value.
    """

    weights = weights/np.sum(weights)
    sorter_sample = np.argsort(samples)

    sorted_weights = weights[sorter_sample]
    sorted_samples = samples[sorter_sample]

    best_val_idx = np.argmin(np.abs(sorted_samples-float(best_val)))

    # Faster method
    best_percentile = np.sum(sorted_weights[:best_val_idx])
    model = weightstats.DescrStatsW(sorted_samples, sorted_weights)
    lower_error, upper_error = model.quantile([(1-.6827)*best_percentile, best_percentile+(.6827*(
        1-best_percentile))], return_pandas=False)-sorted_samples[best_val_idx]

    """
    # Previous method (slower)
    # Shows more detail on how we are getting the errors
    
    total_sum = 0
    for i in range(best_val_idx, len(sorted_weights)):
        total_sum += sorted_weights[i]
        if total_sum >= .6827*np.sum(sorted_weights[best_val_idx:]):
            upper_error = sorted_samples[i]-best_val
            break

    total_sum = 0
    for i in range(best_val_idx, 0, -1):
        total_sum += sorted_weights[i]
        if total_sum >= .6827*np.sum(sorted_weights[:best_val_idx]):
            lower_error = sorted_samples[i]-best_val
            break"""

    return -np.abs(lower_error), np.abs(upper_error)

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~#


def make_dict(val_dict, key, vals):
    """makes dictionary using the given values. if the key already 
    exists, then the values get appended. 

    Args:
        val_dict (dict): the dictionary where the values are to be added
        key (str): the key of the value
        vals (array): the values to be added
    """
    if key in val_dict:
        val_dict[key] = np.ndarray.flatten(np.append(val_dict[key], vals))
    else:
        val_dict[key] = np.ndarray.flatten(vals)


def q_to_u_err(q, q_err):
    """Converts error in q to errors in u parameters for quadratic limb-darkening

    Args:
        q (list): values of q0, q1
        q_err (list): errors on q0, q1

    Returns:
        list: errors on u0, u1
    """

    u_err = []
    u_err.append(np.sqrt(((q[1] * q_err[0])**2)/q[0] + 4 * q[0] * q_err[1]**2))
    u_err.append(np.abs(np.sqrt(
        (((1 - 2 * q[1]) * q_err[0])**2)/(0.25 * q[0]) + 4 * q[0] * q_err[1] ** 2)))

    return u_err


def q_to_u(q, q_err_l, q_err_u):
    """Converts q (Kipping parameters) to limb darkening parameters u.
    Handles only quadratic limb darkening currently.

    Args:
        q (list): values of q0, q1
        q_err_l (list): lower bound of errors on q0, q1
        q_err_u (list): upper bound of errors on q0, q1

    Returns:
        tuple: the values of u, lower bound on values, upper bound on values.
    """

    q_err_l = np.abs(np.array(q_err_l,dtype=float))
    q_err_u = np.abs(np.array(q_err_u,dtype=float))


    u = [2 * np.sqrt(q[0]) * q[1], np.sqrt(q[0]) * (1 - 2 * q[1])]

    u_err_l = q_to_u_err(q, q_err_l)
    u_err_u = q_to_u_err(q, q_err_u)

    return np.array(u,dtype=float), -np.abs(np.array(u_err_l,dtype=float)), np.abs(np.array(u_err_u,dtype=float))


class ErrorLimits:
    """
    Initializes the error limit analysis.

    Args:
        output_parmeters (str): path of the output_parameters folder
                                generated by TransitFit
    """

    def __init__(self, output_parmeters):
        
        # The path to the output_parameters folder from the TransitFit output.
        self.OUTPUT_PARAMETERS_FOLDER = output_parmeters

        # This will be the name of the output file
        self.MODIFIED_SUMMARY_OUTPUT = 'modified_output.csv'

        # Summary_output file
        self.summary_output = glob.glob(
            self.OUTPUT_PARAMETERS_FOLDER+'/*summary_output.csv')
        if len(self.summary_output) == 0:
            #print('')
            raise RuntimeError(
                'Please check the pathname, it may not be correct.')

            # return

        # Searching for all the .pkl files
        self.files_pkl = glob.glob(
            self.OUTPUT_PARAMETERS_FOLDER+'/quicksaves/*results.pkl')
        self.files_pkl.sort()

        # Searching for all the .csv files
        self.output_csvs = glob.glob(
            self.OUTPUT_PARAMETERS_FOLDER+'/quicksaves/*output.csv')
        self.output_csvs.sort()

        # In folded mode, there is additional filter-dependent folder inside
        # output_parameters, checking if that is the case.
        self.folded_mode_csv = glob.glob(
            self.OUTPUT_PARAMETERS_FOLDER+'/*/quicksaves/*output.csv')

        if len(self.folded_mode_csv) != 0:
            print('Folded mode detected.')
            self.FOLDED_MODE = True
            self.folded_mode_pkl = glob.glob(
                self.OUTPUT_PARAMETERS_FOLDER+'/*/quicksaves/*results.pkl')
            self.folded_mode_pkl.sort()
            self.folded_mode_csv.sort()

            self.folded_mode_priors_pkl = glob.glob(
                self.OUTPUT_PARAMETERS_FOLDER+'/*/quicksaves/*priors.pkl')
            self.folded_mode_priors_pkl.sort()
            # if len(unfolded_priors_pkl)>0:
            with open(self.folded_mode_priors_pkl[0], 'rb') as handle:
                priors = pickle.load(handle)

            self.allow_ttv = priors.allow_ttv
            if self.allow_ttv:
                print('TTV mode detected.')

        else:
            self.allow_ttv = False
            self.FOLDED_MODE = False
            self.folded_params = []
            self.priors_pkl = glob.glob(
                self.OUTPUT_PARAMETERS_FOLDER+'/quicksaves/*priors.pkl')
            self.priors_pkl.sort()
            with open(self.priors_pkl[0], 'rb') as handle:
                priors = pickle.load(handle)

        # The list of required parameters to analyze
        self.required_params = ['P', 't0', 'a/r*', 'inc', 'rp/r*']
        self.limb_dark_coeffs = priors.limb_dark_coeffs#['q0','q1']#
        self.ld_method = priors.limb_dark #'quadratic' #
        self.values = {}

    def handle_ttv(self):
        """ Handles the case of TTV analysis.
        """
        all_epochs = np.empty(0)
        for i, fmpk in enumerate(self.folded_mode_priors_pkl):
            with open(fmpk, 'rb') as handle:
                priors = pickle.load(handle)

            fitting_params = priors.fitting_params  # [name, tidx,fidx, eidx]
            # filters = fitting_params[:,2]
            params = fitting_params[:, 0]

            df = pd.read_csv(self.folded_mode_csv[i])
            selected_df = df.loc[df['Parameter'] == 't0']
            epochs = selected_df['Epoch'].to_numpy(dtype=float)
            all_epochs = np.append(all_epochs, epochs)
            with open(self.folded_mode_pkl[i], 'rb') as handle:
                results = pickle.load(handle)

            for j, p in enumerate(params):
                if p == 't0':
                    make_dict(self.values, p+'_' +
                              str(int(epochs[j])), results.samples[:, j])
                    make_dict(self.values, p+'_' +
                              str(int(epochs[j]))+'_weights', results.weights)
                    self.values[p+'_'+str(int(epochs[j])) +
                                '_best'] = results.best[j]

        priors_pkl = glob.glob(
            self.OUTPUT_PARAMETERS_FOLDER+'/quicksaves/*priors.pkl')
        priors_pkl.sort()
        for i, pp in enumerate(priors_pkl):
            with open(pp, 'rb') as handle:
                priors = pickle.load(handle)
            fitting_params = priors.fitting_params  # [name, tidx,fidx, eidx]
            # filters=fitting_params[:,2]
            params = fitting_params[:, 0]

            with open(self.files_pkl[i], 'rb') as handle:
                results = pickle.load(handle)

            for j, p in enumerate(params):
                # elif params[p] in {'a','inc','rp','q0','q1'}:
                make_dict(self.values, p, results.samples[:, j])
                make_dict(self.values, p+'_weights', results.weights)
                self.values[p+'_best'] = results.best[j]

        all_epochs = np.unique(all_epochs)

        # Generating the output
        with open(self.OUTPUT_PARAMETERS_FOLDER+'/'+self.MODIFIED_SUMMARY_OUTPUT, 'w') as mso:
            mso.write(
                'Parameter,Filter,Epoch,Best,Lower_error,Upper_error\n')
            for e in all_epochs:
                param_ = 't0_'+str(int(e))
                errs = get_quantiles_on_best_val(
                    samples=self.values[param_], weights=self.values[param_+'_weights'], best_val=self.values[param_+'_best'])
                # (model.mean)
                mso.write(
                    f"t0,-,{e},{self.values[param_+'_best']},{errs[0]},{errs[1]}\n")

            q, q_low, q_up = [], [], []
            for p in params:
                param_ = p
                errs = get_quantiles_on_best_val(
                    samples=self.values[param_], weights=self.values[param_+'_weights'], best_val=self.values[param_+'_best'])
                # (model.mean)

                mso.write(
                    f"{p},-,-,{self.values[param_+'_best']},{errs[0]},{errs[1]}\n")

                if p in self.limb_dark_coeffs:
                    q.append(self.values[param_+'_best']),
                    q_low.append(errs[0])
                    q_up.append(errs[1])
            
            if self.ld_method=='quadratic':
                u, u_low, u_up = q_to_u(q, q_low, q_up)

                for i, p in enumerate(self.limb_dark_coeffs):
                    mso.write(f"u{str(i)},-,-,{u[i]},{u_low[i]},{u_up[i]}\n")
                    #mso.write(f"u0,-,-,{u[0]},{u_low[0]},{u_up[0]}\n")
                    #mso.write(f"u1,-,-,{u[1]},{u_low[1]},{u_up[1]}\n")

        print(
            f"Saved results in {self.OUTPUT_PARAMETERS_FOLDER+'/'+self.MODIFIED_SUMMARY_OUTPUT}")

        return

    def get_errors(self):
        """ Gets the upper and lower error bound on values.
        """

        # Initalising values dictionary and filters list.
        # selvalues = {}
        filters = []

        if self.allow_ttv:
            self.handle_ttv()
            return

        # Getting the values of 'P' and 't0' in case of folded mode.
        elif self.FOLDED_MODE:
            self.folded_params = self.required_params[:2]
            self.required_params = self.required_params[2:]
            for i, file in enumerate(self.folded_mode_pkl):
                with open(file, 'rb') as handle:
                    result = pickle.load(handle)

                df_csv = pd.read_csv(self.folded_mode_csv[i])

                for i, param in enumerate(self.folded_params):
                    selected_df = df_csv.loc[df_csv['Parameter'] == param]
                    index = selected_df.index

                    make_dict(self.values, param, result.samples[:, index])
                    make_dict(self.values, param+'_weights', result.weights)

        for i, file in enumerate(self.files_pkl):
            with open(file, 'rb') as handle:
                result = pickle.load(handle)

            df_csv = pd.read_csv(self.output_csvs[i])

            for i, param in enumerate(self.required_params):
                selected_df = df_csv.loc[df_csv['Parameter'] == param]
                index = selected_df.index
                if param in {'inc', 'rp/r*'}:
                    # To account for a/AU parameter present in the
                    # summary_output files.
                    index = index-1
                
                if '-' not in selected_df['Filter'].to_list():
                    batch_filters = selected_df['Filter'].to_list()
                    filters += batch_filters
                    for i, f in enumerate(batch_filters):
                        make_dict(self.values, param+'_'+f,
                                  result.samples[:, index[i]])
                        make_dict(self.values, param+'_'+f +
                                  '_weights', result.weights)
                        
                        for j,u in enumerate(self.limb_dark_coeffs):
                            i = len(self.limb_dark_coeffs)*i  
                            # limb darkenening coeffs are saved together in the .pkl files
                            
                            make_dict(self.values, 'q'+str(j)+'_'+f,
                                    result.samples[:, index[-1]+1+i])
                            make_dict(self.values, 'q'+str(j)+'_'+f +
                                    '_weights', result.weights)
                            
                            #if self.ld_method=='quadratic':
                            """make_dict(self.values, 'q0'+'_'+f,
                                    result.samples[:, index[-1]+1+i])
                            make_dict(self.values, 'q1'+'_'+f,
                                    result.samples[:, index[-1]+1+i+1])

                            make_dict(self.values, 'q0'+'_'+f +
                                    '_weights', result.weights)
                            make_dict(self.values, 'q1'+'_'+f +
                                    '_weights', result.weights)"""
                
                elif selected_df['Error'].iloc[0]=='-':
                    self.required_params.remove(param)
                
                else:

                    make_dict(self.values, param, result.samples[:, index])
                    make_dict(self.values, param+'_weights', result.weights)

                    if param == 'rp/r*':  # Folded mode
                        for j,u in enumerate(self.limb_dark_coeffs):

                            make_dict(self.values, 'q'+str(j),
                                    result.samples[:, index+1])
                            make_dict(self.values, 'q'+str(j)+'_weights', result.weights)

                            """make_dict(self.values, 'q1',
                                    result.samples[:, index+2])

                            make_dict(self.values, 'q'+str(j)+'_weights', result.weights)
                            make_dict(self.values, 'q1'+'_weights', result.weights)"""
        
        #if self.ld_method=='quadratic':
        self.required_params = self.folded_params + \
            self.required_params+self.limb_dark_coeffs#['q0', 'q1']
            

        filters = list(set(filters))
        filters = np.sort(np.array(filters, dtype=int))

        df_so = pd.read_csv(self.summary_output[0])

        # Getting the best_values from the summary_output.csv file
        for p in self.required_params:
            selected_df = df_so.loc[df_so['Parameter'] == p]
            batch_filters = selected_df['Filter']
            if len(selected_df) == 1 and '-' not in batch_filters:
                self.values[p+'_best'] = float(selected_df['Best'])

            else:
                #batch_filters = selected_df['Filter']
                bestvals = selected_df['Best'].to_numpy(dtype=float)
                for i, f in enumerate(batch_filters):
                    self.values[p+'_'+str(f)+'_best'] = float(bestvals[i])

        # Generating the output
        with open(self.OUTPUT_PARAMETERS_FOLDER+'/'+self.MODIFIED_SUMMARY_OUTPUT, 'w') as mso:
            mso.write('Parameter,Filter,Best,Lower_error,Upper_error\n')
            q, q_low, q_up = [], [], []
            q_dict = {}
            #print(self.values)

            for param in self.required_params:
                if param in ['rp/r*']+self.limb_dark_coeffs and len(filters) > 0:
                    for fil in filters:
                        param_ = param+'_'+str(int(fil))
                        try:
                            best_val=self.values[param_+'_best']
                            #errs = get_quantiles_on_best_val(
                            #    samples=self.values[param_], weights=self.values[param_+'_weights'], best_val=self.values[param_+'_best'])
                        except KeyError:
                            best_val=self.values[param+'_best']
                        
                        errs = get_quantiles_on_best_val(
                            samples=self.values[param_], weights=self.values[param_+'_weights'], best_val=best_val)

                        # (model.mean)
                        mso.write(
                            f"{param},{fil},{best_val},{errs[0]},{errs[1]}\n")

                else:
                    errs = get_quantiles_on_best_val(
                        samples=self.values[param], weights=self.values[param+'_weights'], best_val=self.values[param+'_best'])
                    mso.write(
                        f"{param},-,{self.values[param+'_best']},{errs[0]},{errs[1]}\n")

                    if param in self.limb_dark_coeffs:
                        q.append(self.values[param+'_best']),
                        q_low.append(errs[0])
                        q_up.append(errs[1])

            if self.ld_method=='quadratic':
                if len(q) > 0:
                    u, u_low, u_up = q_to_u(q, q_low, q_up)
                    
                    for j,p in enumerate(self.limb_dark_coeffs):
                        mso.write(f"u{str(j)},-,{u[j]},{u_low[j]},{u_up[j]}\n")
                        #mso.write(f"u1,-,{u[1]},{u_low[1]},{u_up[1]}\n")

                elif len(filters) > 0:
                    for fil in filters:
                        q, q_low, q_up = [], [], []
                        for param in self.limb_dark_coeffs:
                            param_ = param+'_'+str(int(fil))
                            try:
                                best_val=self.values[param_+'_best']
                            except:
                                best_val=self.values[param+'_best']
                            errs = get_quantiles_on_best_val(
                                samples=self.values[param_], weights=self.values[param_+'_weights'], best_val=best_val)

                            q.append(best_val),
                            q_low.append(errs[0])
                            q_up.append(errs[1])

                        u, u_low, u_up = q_to_u(q, q_low, q_up)

                        for j,p in enumerate(self.limb_dark_coeffs):
                            mso.write(
                                f"u{str(j)},{str(int(fil))},{u[j]},{u_low[j]},{u_up[j]}\n")
                            #mso.write(f"u1,{str(int(fil))},{u[1]},{u_low[1]},{u_up[1]}\n")

        print(
            f"Saved results in {self.OUTPUT_PARAMETERS_FOLDER+'/'+self.MODIFIED_SUMMARY_OUTPUT}")

        return



'''
Object to handle and deal with prior info for retrieval
'''

import numpy as np
from collections.abc import Iterable

from ._params import _Param, _UniformParam, _GaussianParam
from .detrending_funcs import NthOrderDetrendingFunction
from .detrender import DetrendingFunction
from ._limb_darkening_handler import LimbDarkeningHandler
from ._paramarray import ParamArray

# Parameters which do not vary with telescope, filter, or epoch
# Note that t0 is in here, as it is global unless ttv mode is on
_global_params = ['P', 'a', 'inc', 'ecc', 'w', 't0']

# Parameters which vary with filter
_filter_dependent_params = ['rp', 'q0', 'q1', 'q2', 'q3']

# Params which vary with epoch
_epoch_dependent_params = []

# The possible ldc lables
_all_ldcs = ['q0', 'q1', 'q2', 'q3']

# Default values for parameters (used in io.parse_priors_list)
_prior_info_defaults = {'P':1, 'a':10, 'inc':90, 'rp':0.05, 't0':0, 'ecc':0,
                        'w':90, 'limb_dark':'quadratic', 'q0':0.1, 'q1':0.3,
                        'q2':None, 'q3':None, 'n_telescopes':1,  'n_filters':1,
                        'n_epochs':1, 'norm':1,'escale':0}

def setup_priors(P, t0, a, rp, inc, ecc, w, limb_dark, n_telescopes, n_filters,
                 n_epochs, q0=None, q1=None, q2=None, q3=None, allow_ttv=False,
                 lightcurves=None, error_scaling=False):
    '''
    Factory function to initialise a PriorInfo object

    qX should either be a single value or a list of values length n_filters
    '''
    # Deal with the qX values
    if q0 is not None and not isinstance(q0, Iterable):
        q0 = [q0 for i in range(n_filters)]
    if q1 is not None and not isinstance(q1, Iterable):
        q1 = [q1 for i in range(n_filters)]
    if q2 is not None and not isinstance(q2, Iterable):
        q2 = [q2 for i in range(n_filters)]
    if q3 is not None and not isinstance(q3, Iterable):
        q3 = [q3 for i in range(n_filters)]

    default_dict = {'P':P, 't0':t0, 'a':a, 'rp':rp, 'inc':inc, 'ecc':ecc,
                    'w':w, 'q0':q0, 'q1':q1, 'q2':q2, 'q3':q3, 'norm':1}#,'escale':0}

    if error_scaling:
        default_dict = {'P':P, 't0':t0, 'a':a, 'rp':rp, 'inc':inc, 'ecc':ecc,
                    'w':w, 'q0':q0, 'q1':q1, 'q2':q2, 'q3':q3, 'norm':1,'escale':0}

    return PriorInfo(default_dict, limb_dark, n_telescopes, n_filters,
                      n_epochs, allow_ttv, lightcurves)



class PriorInfo:
    '''
    Object to deal with anything involving priors

    Parameters
    ----------
    default_dict : dictionary
        Dictionary containing default parameter values.
    limb_dark : str
        The limb darkening model to use.
    n_telescopes : int
        The number of different telescopes used in the data set
    n_filters : int
        The number of different filters used in the data set
    n_epochs : int
        The number of different epochs used in the data set
    allow_ttv : bool, optional
        If True, will fit t0 to each transit. Default is False.
    lightcurves : array_like, shape(n_telescopes, n_filters, n_epochs)
        Array of the light curves. If there is no observation for a particular
        telescope, filter, epoch combination, then the value should be set to
        None.
    '''
    def __init__(self, default_dict, limb_dark, n_telescopes, n_filters,
                 n_epochs, allow_ttv=False, lightcurves=None):
        # Store the basics
        self.limb_dark = limb_dark
        self.n_telescopes = n_telescopes
        self.n_filters = n_filters
        self.n_epochs = n_epochs
        self.allow_ttv = allow_ttv

        # Initialise limb darkening (set to off)
        self.ld_handler = LimbDarkeningHandler(self.limb_dark)
        self.limb_dark_coeffs = self.ld_handler.get_required_coefficients()
        self.fit_ld = False
        self.ld_fit_method = 'off'

        # Initialse detrending (set to off)
        self.detrend = False
        self.detrending_coeffs = []

        # Initialse normalisation (set to off)
        self.normalise=False

        self.error_scaling = False

        #####################
        # Set up the priors #
        #####################
        # Dictionary containing info on all the priors
        self.priors = {}

        for key in default_dict.keys():
            #print('Param:', key)
            # Loop through each parameter and initialise the prior


            if key in _global_params:
                # These are global parameters
                if key == 't0' and allow_ttv:
                    # We are fitting a separate t0 for each epoch
                    self.priors[key] = ParamArray(key, (1,1, self.n_epochs), False, False, True, default_dict[key])
                else:
                    self.priors[key] = ParamArray(key, (1,1, 1), False, False, False, default_dict[key])

            elif key in _filter_dependent_params:
                # These parameters vary with filter
                shape = (1, self.n_filters, 1)

                # Deal with the ldcs
                if key in _all_ldcs:
                    if key in self.limb_dark_coeffs:
                        #print('Initialising ldc', key)
                        self.priors[key] = ParamArray(key, shape, False, True, False, default_dict[key][0])
                        for qi, q in enumerate(default_dict[key]):
                            self.priors[key].set_value(q, filter_idx=qi)

                # Otherwise, this is rp
                else:
                    self.priors[key] = ParamArray(key, shape, False, True, False, default_dict[key])

            else:
                # This is a lightcurve dependent parameter
                self.priors[key] = ParamArray(key, (self.n_telescopes, self.n_filters,self.n_epochs), True, True, True, default_dict[key], lightcurves=lightcurves)

        # All priors should now be initialised

        # We also need a list for keeping track of what is being fitted
        # Each entry in this will be
        # [param name, telescope index, filter index, epoch index]
        self.fitting_params = None

        #for p in self.priors.keys():
        #    print(p, self.priors[p].array)

    ###############################################################
    #                   ADDING FIT PARAMS                         #
    ###############################################################

    def add_uniform_fit_param(self, name, low_lim, high_lim,
                              telescope_idx=None, filter_idx=None, epoch_idx=None):
        '''
        Adds a new parameter which will be fitted uniformly in the range given
        by low_lim and high_lim
        '''
        if name in ['a', 'P', 'rp', 'inc', 'ecc', 'w']:
            negative_allowed = False
        else:
            negative_allowed = True
        
        if name in ['inc']:
            high_lim = 90
        
        #if name not in self.priors and name=='escale':
        #    # Need to initialise the entry in the priors dict
        #    self.priors[name] = ParamArray(name, (self.n_telescopes, self.n_filters,self.n_epochs), True, True, True)

        self.priors[name].add_uniform_fit_param(low_lim, high_lim,
                                                telescope_idx, filter_idx,
                                                epoch_idx, negative_allowed)

        # Store some info for later
        if self.fitting_params is None:
            self.fitting_params = np.array([[name, telescope_idx, filter_idx, epoch_idx]], dtype=object)
        else:
            self.fitting_params = np.append(self.fitting_params, np.array([[name, telescope_idx, filter_idx, epoch_idx]], object), axis=0)
        #self.fitting_params.append([name, telescope_idx, filter_idx, epoch_idx])

    def add_gaussian_fit_param(self, name, mean, stdev,
                               telescope_idx=None, filter_idx=None, epoch_idx=None):
        '''
        Adds a new parameter which will be fitted with a Gaussian prior
        '''
        # If this is t0 and ttvs are on, we want to make the prior much wider to allow for proper TTVs
        #if name == 't0' and self.allow_ttv:
            # Set the Gaussian width to 0.1 days for ttv fitting mode
            #stdev = 0.1
        if name in ['a', 'P', 'rp', 'inc', 'ecc', 'w']:
            negative_allowed = False
        else:
            negative_allowed = True
        
        if name in 'inc':
            clipped_gaussian=True
        else:
            clipped_gaussian=False

        self.priors[name].add_gaussian_fit_param(mean, stdev,
                                                 telescope_idx, filter_idx,
                                                 epoch_idx, negative_allowed, clipped_gaussian)

        # Store some info for later
        if self.fitting_params is None:
            self.fitting_params = np.array([[name, telescope_idx, filter_idx, epoch_idx]], dtype=object)
        else:
            self.fitting_params = np.append(self.fitting_params, np.array([[name, telescope_idx, filter_idx, epoch_idx]], object), axis=0)


    ###############################################################
    #         ADDING DETRENDING/NORMALISATION/LDC FITTING         #
    ###############################################################

    def fit_detrending(self, lightcurves, method_list, method_index_array,
                       limits=None):
        '''
        Intialises detrending
        '''
        if self.detrend:
            raise ValueError('Detrending is already initialised. You need to make a new PriorInfo to use another detrending method!')

        # Store some info - used in splitting
        self._detrend_method_list = method_list
        self._detrend_method_index_array = method_index_array

        self.detrending_coeffs = [[] for method in method_list]

        # Go through each lightcurve and initialise the relevant detrending
        for i in np.ndindex(lightcurves.shape):
            telescope_idx, filter_idx, epoch_idx = i

            if lightcurves[i] is not None:
                # We have a light curve, pull out the method info
                method_idx = method_index_array[i]
                method = method_list[method_idx]

                if method[0] == 'off':
                    # No detrending - skip this curve
                    pass
                else:
                    # First we set up the detrending for the lightcurve
                    if method[0] == 'nth order':
                        lightcurves[i].set_detrending(method[0], order=method[1], method_idx=method_idx)
                    elif method[0] == 'custom':
                        lightcurves[i].set_detrending(method[0], function=method[1], method_idx=method_idx)

                    # Now set up the required parameters to be fitted
                    n_coeffs = lightcurves[i].n_detrending_params

                    for coeff_i in range(n_coeffs):
                        # Loop over each coefficient!
                        coeff_name = 'd{}_{}'.format(coeff_i, method_idx)

                        # Set up the limits of each fitting parameter
                        if limits is None:
                            # Limits are not provided.
                            # We hard-code a default of ±10
                            low_lim = -10.
                            high_lim = 10.

                        elif isinstance(limits[method_idx][0], Iterable):
                            # The limits for each parameter have been set
                            # individually
                            low_lim = limits[method_idx][coeff_i][0]
                            high_lim = limits[method_idx][coeff_i][1]

                        else:
                            # [low, high] limit provided to apply universally
                            # to all the coefs for this method.
                            low_lim = limits[method_idx][0]
                            high_lim = limits[method_idx][1]

                        if method[0] == 'custom':
                            # Work out telescope, filter, epoch dependencies
                            tel_dep = coeff_i + 1 in method[2]
                            filt_dep = coeff_i + 1 in method[3]
                            epoch_dep = coeff_i + 1 in method[4]
                        else:
                            # We assume that the detrending coefficients are
                            # independent for all lightcurves
                            # (i.e. all dependencies are True)
                            tel_dep = True
                            filt_dep = True
                            epoch_dep = True

                        # Change the indices to None if not dependent
                        if tel_dep:
                            coeff_telescope_idx = telescope_idx
                        else:
                            coeff_telescope_idx = None
                        if  filt_dep:
                            coeff_filter_idx = filter_idx
                        else:
                            coeff_filter_idx = None
                        if epoch_dep:
                            coeff_epoch_idx = epoch_idx
                        else:
                            coeff_epoch_idx = None

                        if coeff_name not in self.priors:
                            # Need to initialise the entry in the priors dict
                            self.detrending_coeffs[method_idx].append(coeff_name)

                            # Have to deal with nth order and custom separately
                            if method[0] == 'nth order':
                                shape = (self.n_telescopes, self.n_filters, self.n_epochs)

                                # Initialise the prior
                                self.priors[coeff_name] = ParamArray(coeff_name, shape, True, True, True)

                            elif method[0] == 'custom':
                                # Work out the shape needed for the ParamArray
                                shape = ([1, self.n_telescopes][tel_dep],
                                         [1, self.n_filters][filt_dep],
                                         [1, self.n_epochs][epoch_dep])

                                # Initialise the prior
                                self.priors[coeff_name] = ParamArray(coeff_name, shape, tel_dep, filt_dep, epoch_dep)


                        # Now set up the fitting. We assume a uniform prior
                        # We have to check here for parameters which are
                        # telescope/filter/epoch dependent!
                        if coeff_name not in self.fitting_params[:,0]:
                            # The parameter has not been set up for fitting
                            # -> no checks needed
                            self.add_uniform_fit_param(coeff_name, low_lim, high_lim, coeff_telescope_idx, coeff_filter_idx, coeff_epoch_idx)


                        else:
                            # Run a check to see if the param is at least one
                            # of telescope, filter, or epoch dependent. If so,
                            # this needs to be fitted, but otherwise ignored.
                            if tel_dep or filt_dep or epoch_dep:
                                self.add_uniform_fit_param(coeff_name, low_lim, high_lim, coeff_telescope_idx, coeff_filter_idx, coeff_epoch_idx)

        self.detrend=True

    def fit_limb_darkening(self, fit_method='independent',
                           host_T=None, host_logg=None,
                           host_z=None, filters=None,
                           n_samples=20000, do_mc=False, cache_path=None,
                           ldtk_uncertainty_multiplier=1.):
        '''
        Initialises fitting of limb darkening parameters, either independently
        or coupled across wavebands.

        Parameters
        ----------
        fit_method : str, optional
            Determines the mode of fitting of LD parameters. The available
            modes are:
                - 'coupled' : all limb darkening parameters are fitted
                  independently, but are coupled to a wavelength dependent
                  model based on the host parameters
                - 'single' : LD parameters are still tied to a model, but only
                  the first filter is actively fitted. The remaining filters
                  are estimated based off the ratios given by ldtk for a host
                  with the given parameters. This mode is useful for a large
                  number of filters, as 'coupled' or 'independent' fitting will
                  lead to much higher computation times.
                - 'independent' : Each LD coefficient is fitted separately for
                  each filter, with no coupling to the ldtk models.
            The default is 'independent'.
        host_T : tuple or None, optional
            The effective temperature of the host star, in Kelvin, given as a
            (value, uncertainty) pair. Must be provided if fit_method is
            'coupled' or 'single'. Default is None.
        host_logg : tuple or None, optionalor None, optional
            The log_10 of the surface gravity of the host star, with gravity
            measured in cm/s2. Should be given as a (value, uncertainty) pair.
            Must be provided if fit_method is 'coupled' or 'single'. Default is
            None.
        host_z : tuple or None, optional
            The metalicity of the host, given as a (value, uncertainty) pair.
            Must be provided if fit_method is 'coupled' or 'single'. Default is
            None.
        filters : array_like or None, optional
            The set of filters, given in [low, high] limits for the wavelengths
            with the wavelengths given in nanometers. The ordering of the
            filters should correspond to the filter_idx parameter used
            elsewhere. Must be provided if fit_method is 'coupled' or 'single'.
            Default is None.
        n_samples : int, optional
            The number of limb darkening profiles to create. Passed to
            ldtk.LDPSetCreator.create_profiles(). Default is 20000.
        do_mc : bool, optional
            If True, will use MCMC to estimate coefficient uncertainties more
            accurately. Default is False.
        cache_path : str, optional
            This is the path to cache LDTK files to. If not specified, will
            default to the LDTK default
        ldtk_uncertainty_multiplier: float, optional
            (From LDTK:) The uncertainty multiplier ϵ is a subjective factor
            that defines how strongly the LD profile (or the prior created
            from it) constrains the final analysis (that is, how much we
            trust the stellar atmosphere models used to create the profiles.)
        '''
        # Sanity checks
        if fit_method not in ['coupled', 'single', 'independent']:
            raise ValueError('Unrecognised fit method {}'.format(fit_method))

        if not fit_method == 'independent':
            if filters is None:
                raise ValueError('Filters must be provided for coupled and single ld_fit_methods')
            if not len(filters) == self.n_filters:
                raise ValueError('{} filters were given, but there are {} filters required!'.format(len(filters), self.n_filters))

        # Store some useful information
        self.fit_ld = True
        self.ld_fit_method = fit_method
        self.filters = filters
        self._n_ld_samples = n_samples
        self._do_ld_mc = do_mc
        self._ld_cache_path = cache_path

        # Set up fitting for each LDC
        for i in range(self.n_filters):
            for ldc, name in enumerate(self.limb_dark_coeffs):
                self.add_uniform_fit_param(name, 0, 1, filter_idx=i)

            # If we are in 'single' mode, we only need to fit for the first
            # wavelength
            if self.ld_fit_method == 'single':
                break

        if not fit_method == 'independent':
            # Now if we are coupling across wavelength we must initialise PyLDTK
            self.ld_handler.initialise_ldtk(host_T, host_logg, host_z, filters,
                                            self.limb_dark,
                                            n_samples, do_mc, cache_path,
                                            ldtk_uncertainty_multiplier)

    def fit_normalisation(self, lightcurves,normalise_limits):
        '''
        When run, the Retriever will fit normalisation of the data as a
        free parameter.

        Parameters
        ----------
        lightcurves : np.array of `LightCurve`s, shape (n_telescopes, n_filters, n_epochs)
            The array of LightCurves to be normalised. We can use the value of
            fluxes to estimate a normalisation constant (c_n) range for each
            light curve. We use
                ``1/f_median -1 <= c_n <= 1/f_median + 1``
            as the default range, where f_median is the median flux value.
        '''
        if self.normalise:
            raise ValueError('Detrending is already initialised. You need to make a new PriorInfo to use another detrending method!')

        for i in np.ndindex(lightcurves.shape):
            telescope_idx, filter_idx, epoch_idx = i

            if lightcurves[i] is not None:
                # A light curve exists. Set up normalisation
                best, low, high = lightcurves[i].set_normalisation(normalise_limits)
                self.add_uniform_fit_param('norm', low, high, telescope_idx, filter_idx, epoch_idx)

        self.normalise = True

    def set_error_scaling(self,lightcurves, scaling_limits, method_index_array):#, telescope_idx=None):
        
        for i in np.ndindex(lightcurves.shape):
            telescope_idx, filter_idx, epoch_idx = i

            if lightcurves[i] is not None:
                # We have a light curve, pull out the method info
                method_idx = method_index_array[i]

                if isinstance(scaling_limits[0], Iterable):
                    # The limits for each parameter have been set
                    # individually
                    #low_lim = scaling_limits[telescope_idx][0]
                    #high_lim = scaling_limits[telescope_idx][1]
                    scaling_limits_iter = scaling_limits[method_idx]

                else:
                    scaling_limits_iter = scaling_limits

                if lightcurves[i] is not None:
                    # A light curve exists. Set up error scaling
                    best, low, high = lightcurves[i].set_error_scaling(scaling_limits_iter)
                    self.add_uniform_fit_param('escale', low, high, telescope_idx, filter_idx, epoch_idx)

        self.error_scaling = True
        #key='escale'
        #self.priors[key] = ParamArray(key, (self.n_telescopes, self.n_filters,self.n_epochs), True, True, True, 0, lightcurves=lightcurves)



    ###############################################################
    #             CONVERSIONS FOR FITTING ROUTINES                #
    ###############################################################

    def _convert_unit_cube(self, cube):
        '''
        Takes the unit cube provided by dynesty (all values between 0 and 1)
        and converts them into physical values
        '''
        new_cube = np.zeros(len(self.fitting_params))

        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info

            if name == 'inc':
                idx = self.priors[name]._generate_idx(tidx, fidx, eidx)

                if hasattr(self.priors[name].array[idx], "uniform_to_clipped_gaussian"):
                    new_cube[i] = self.priors[name].clipped_gaussian_transform(cube[i], tidx, fidx, eidx)

                else:
                    new_cube[i] = self.priors[name].from_unit_interval(cube[i], tidx, fidx, eidx)

            else:
                new_cube[i] = self.priors[name].from_unit_interval(cube[i], tidx, fidx, eidx)

        return new_cube

    def _interpret_param_array(self, array):
        '''
        Interprets the parameter cube generated by dynesty and returns them in
        a format usable by the LikelihoodCalculator
        '''

        # Initialse a dictionary to store the physical results in
        result = {}

        for name in self.priors.keys():
            # First we initialise the entry in the results dictionary
            result[name] = self.priors[name].generate_blank_ParamArray()

        initialised_params = []
        # Now we add in the results
        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info
            result[name][tidx, fidx, eidx] = array[i]
            initialised_params.append(name)

        # If things aren't being fitted, add these in too
        for name in result.keys():
            if name not in initialised_params:
                for i in np.ndindex(self.priors[name].shape):
                    result[name][i] = self.priors[name][i]

        return result

    def _interpret_final_results(self, results):
        '''
        Generates a dictionary of results and a dictionary of errors from the
        final results of a run
        '''
        best_results = results.best
        best_result_errors = results.uncertainties

        result_dict = {}
        errors_dict = {}

        # Initialise the entries in the dictionaries
        for name in self.priors.keys():
            # First we initialise the entry in the results dictionary
            result_dict[name] = self.priors[name].generate_blank_ParamArray()
            errors_dict[name] = self.priors[name].generate_blank_ParamArray()

        initialised_params = []
        # Go through the results object and add in the results - only includes
        # fitted parameters
        for i, param_info in enumerate(self.fitting_params):
            name, tidx, fidx, eidx = param_info
            result_dict[name][tidx, fidx, eidx] = best_results[i]
            errors_dict[name][tidx, fidx, eidx] = best_result_errors[i]
            initialised_params.append(name)

        # Now go through and fill in any non-fitted parameters
        for name in result_dict.keys():
            if name not in initialised_params:
                for i in np.ndindex(self.priors[name].shape):
                    result_dict[name][i] = self.priors[name][i]
                    errors_dict[name][i] = 0

        return result_dict, errors_dict


    ###############################################################
    #                          MISC                               #
    ###############################################################
    def __str__(self):
        print_str = 'Priors:\n'
        print_str += 'Limb darkening model: {}\n'.format(self.limb_dark)
        print_str += 'n telescopes: {}\n'.format(self.n_telescopes)
        print_str += 'n filters: {}\n'.format(self.n_filters)
        print_str += 'n epochs: {}\n'.format(self.n_epochs)
        for var in self.priors:
            print_str += self.priors[var].__str__()
        print_str += 'Total {} fitting parameters'.format(len(self.fitting_params))
        return print_str
    
    def get_latex_friendly_labels(self):
        """Adds a $ around the underscored text to make the latex displays work"""
        labels = []
        for param in self.fitting_params[:,0]:
            if "_" in param:
                new_label = ''
                for s in param.split("_")[:-1]:
                    new_label += s + "$_"
                new_label += param.split("_")[-1]+"$"
            elif param[-1].isdigit() or param =='rp':
                new_label = param[:-1]+'$_'+param[-1]+'$'
            else:
                new_label = param
            labels.append(new_label)
        return labels



'''
Module to calculate the likelihood of a set of parameters
'''

import numpy as np
import batman
from copy import deepcopy
from ._paramarray import ParamArray
#from collections.abc import Iterable



class LikelihoodCalculator:
    '''
    Object to quickly calculate the likelihood of a set of parameters to
    fit a given set of light curves.

    Parameters
    ----------
    lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
        An array of LightCurves. If no data exists for a point in the array
        then the entry should be `None`.
    priors : PriorInfo
        The PriorInfo object for retrieval
    '''
    def __init__(self, lightcurves, priors):
        lightcurves = deepcopy(lightcurves)
        self.lightcurves = np.array(lightcurves, dtype=object)

        self.n_telescopes = self.lightcurves.shape[0]
        self.n_filters = self.lightcurves.shape[1]
        self.n_epochs = self.lightcurves.shape[2]

        self.num_light_curves = len(np.where(self.lightcurves.flatten() != None)[0])

        self.priors = priors

        # We need to make a separate TransitParams and TransitModels for each
        # light curve.

        # Initialise them:
        self.batman_params = ParamArray('batman_params', (self.n_telescopes, self.n_filters, self.n_epochs), True, True, True, lightcurves=self.lightcurves)
        self.batman_models = ParamArray('batman_models', (self.n_telescopes, self.n_filters, self.n_epochs), True, True, True, lightcurves=self.lightcurves)


        for i in np.ndindex(self.lightcurves.shape):
            tidx, fidx, eidx = i

            if self.lightcurves[i] is not None:
                # Set up the params
                self.batman_params.set_value(batman.TransitParams(), tidx, fidx, eidx)

                # Set up the TransitModels
                # Make some realistic parameters to setup the models with
                default_params = batman.TransitParams()
                if self.priors.allow_ttv:
                    default_params.t0 = priors.priors['t0'].default_value
                else:
                    default_params.t0 = priors.priors['t0'].default_value
                default_params.per = priors.priors['P'].default_value
                default_params.rp = priors.priors['rp'].default_value
                default_params.a = priors.priors['a'].default_value
                default_params.inc = priors.priors['inc'].default_value
                default_params.ecc = priors.priors['ecc'].default_value
                default_params.w = priors.priors['w'].default_value
                default_params.limb_dark = priors.limb_dark
                # Note that technically this is q, not u, but it doesn't matter
                # as we are only initialising here
                default_params.u = [priors.priors[qX].default_value for qX in priors.limb_dark_coeffs]

                # Now make the models
                model = batman.TransitModel(default_params, self.lightcurves[i].times)
                self.batman_models.set_value(model, tidx, fidx, eidx)
    
    # New function to handle multiple processes in the same batch.  
    def find_likelihood_parallel_processed(self, cube):
            priors = self.priors
            params = priors._interpret_param_array(cube)

            ln_likelihood = self.find_likelihood(params)
            
            if priors.fit_ld and not priors.ld_fit_method == "independent":
                # Pull out the q values and convert them
                u = []
                for fi in range(priors.n_filters):
                    q = [params[qX][0, fi, 0] for qX in priors.limb_dark_coeffs]
                    u.append(priors.ld_handler.convert_qtou(*q))
                return ln_likelihood + priors.ld_handler.ldtk_lnlike(
                    u, priors.limb_dark
                )
            else:
                return ln_likelihood

    def find_likelihood(self, params):
        '''
        Finds the likelihood of a set of parameters
        '''
        all_chi2 = []
        n_data_points = 0
        total_chi2 = 0

        for i in np.ndindex(self.lightcurves.shape):
            tidx, fidx, eidx = i

            if self.lightcurves[i] is not None:

                # GENERATE THE MODEL LIGHT CURVE

                # Convert the LDC q values to u:
                u = self.priors.ld_handler.convert_qtou(*[params[qX][i] for qX in self.priors.limb_dark_coeffs])

                # Need to update the parameters
                self.update_params(tidx, fidx, eidx,
                                   params['t0'][i],
                                   params['P'][i],
                                   params['rp'][i],
                                   params['a'][i],
                                   params['inc'][i],
                                   params['ecc'][i],
                                   params['w'][i],
                                   self.priors.limb_dark,
                                   u)

                # Now we calculate the model transits
                model = self.batman_models[i]
                model_flux = model.light_curve(self.batman_params[i])

                # DETREND AND NORMALISE THE DATA TO COMPARE TO THE MODEL
                if self.priors.detrend:
                    # We have to work out the detrending info
                    # Get the method index
                    detrending_index = self.priors._detrend_method_index_array[i]
                    # Now get the parameter values
                    d = [params[di][i] for di in self.priors.detrending_coeffs[detrending_index]]
                    
                else:
                    d = None
                                    
                try:
                    d_new=[*d,params['t0'][i],params['P'][i]]
                except TypeError:
                    d_new=[d,params['t0'][i],params['P'][i]]

                if self.priors.normalise:
                    norm = params['norm'][i]
                else:
                    norm = 1
                detrended_flux, err = self.lightcurves[i].detrend_flux(d_new, norm)
                #detrended_flux, err = self.lightcurves[i].detrend_flux(d, norm)

                #print('Likelihood function detrended flux:')
                #print(detrended_flux)

                # Check that there is actually a transit in the model
                # otherwise we impose a large penalty to the chi2 value
                # This avoids a situation where the models try
                # to completely flatten the light curves, which is wrong!
                if np.isclose(model_flux, 1, atol=0., rtol = 1.e-7).all():
                    total_chi2 += 1e7
                    # Ensures that transit models with depths down to 1e-7 are not penalised. 
                    # This allows for the fact that folded data could be quite sensitive!
                else:
                    # Work out the chi2
                    #chi2 = sum((model_flux - detrended_flux)**2 / err**2)

                    if self.priors.error_scaling:
                        # Scaling the errorbars. https://emcee.readthedocs.io/en/stable/tutorials/line/
                        #if isinstance(params['escale'], Iterable):
                        fn = params['escale'][i]
                        sn_sq = np.power(err,2) + np.power(fn*model_flux,2)
                        total_chi2 += np.sum((np.power((model_flux - detrended_flux),2) / sn_sq) + np.log(2*np.pi*sn_sq))
                    
                    else:
                        total_chi2 += np.sum((model_flux - detrended_flux)**2 / err**2)

                #all_chi2.append(chi2)

        #print(params)
        #print(- total_chi2)
        #return - sum(all_chi2)
        return - total_chi2

    def update_params(self, telescope_idx, filter_idx, epoch_index, t0=None,
                      P=None, rp=None, a=None, inc=None, ecc=None, w=None,
                      limb_dark=None, u=None):
        '''
        Updates self.params with values given
        '''
        if t0 is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].t0 = t0
        if P is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].per = P
        if rp is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].rp = rp
        if a is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].a = a
        if inc is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].inc = inc
        if ecc is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].ecc = ecc
        if w is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].w = w
        if limb_dark is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].limb_dark = limb_dark
        if u is not None:
            self.batman_params.array[telescope_idx, filter_idx, epoch_index].u = u



'''
time_conversions

Module which uses astropy.time to convert between different time standards.
'''

from astropy import time, coordinates as coords, units as u

def MJD_to_BJD(mjd_times, ra, dec, lat, lon, elevation, ra_unit='hourangle'):
    '''
    Takes an array of times in MJD and converts to BJD, using observation and
    telescope data.

    Parameters
    ----------
    mjd_times : array_like, shape (N, )
        The times to be converted from MJD
    ra : tuple, length 3
        The right ascension of the observation in either (d, m, s) or (h, m, s)
        depending on if unit is 'deg' or 'hourangle'
    dec : tuple, length 3
        The declination of the observation in (d, m, s)
    lat : float or str
        The latitude of the observer in degrees or "dd:mm:ss"
    log : float or str
        The longitude of the observer in degrees or "dd:mm:ss"
    elevation : float
        The elevation of the observer in meters
    unit : str, optional
        The unit being used for ra. Either `'deg'` for degrees or `'hourangle'`
        for hour angle. Default is `'hourangle'`.

    Returns
    -------
    bjd_times : array_like, shape (N, )
        The times in Barycentric Julian Date
    '''
    if ra_unit.lower() == 'deg':
        ra_unit = u.deg
    elif ra_unit.lower() == 'hourangle':
        ra_unit = u.hourangle
    else:
        raise ValueError('Unrecognised unit {}'.format(unit))

    ra = coords.Angle(ra, unit=unit)
    dec = coords.Angle(dec, unit=u.deg)

    sky_coord = coords.SkyCoord(ra=ra, dec=dec, frame='icrs')

    obs_location = coords.EarthLocation(lat=lat, lon=lon, height=elevation)

    mjd_times = time.Time(mjd_times, format='mjd', location=obs_location,
                           scale='utc')

    delta_time = mjd_times.light_travel_time(sky_coord)

    bjd_times = mjd_times.tdb + delta_time

    return bjd_times.value




import numpy as np
from ._params import _Param, _UniformParam, _GaussianParam


class ParamArray:
    '''
    The _ParamArray is designed to handle parameters which are being fitted
    by TransitFit. It can deal with parameters which are being fitted over
    different combinations of telescope, filter, and epoch

    If a parameter is lightcurve specific, providing lightcurves will
    ensure that only parameters where a lightcurve exists will be
    initialised

    Parameters
    ----------
    name : str
        The parameter name
    shape : tuple
        The shape of the array to use
    telescope_dependent : bool
        Set to True if the parameter varies across telescopes
    filter_dependent : bool
        Set to True if the parameter varies across filters
    epoch_dependent : bool
        Set to True if the parameter varies across epoch
    default_value : float
        The default value for the parameter
    lightcurves : array-like, shape(shape), optional
        The array of relevant light curves.
    '''
    def __init__(self, name, shape, telescope_dependent, filter_dependent,
                 epoch_dependent, default_value=None, lightcurves=None):

        # Store some useful data
        self.name = name
        self.shape = shape

        self.telescope_dependent = telescope_dependent
        self.filter_dependent = filter_dependent
        self.epoch_dependent = epoch_dependent
        self.global_param = sum((self.telescope_dependent,
                                self.filter_dependent, self.epoch_dependent)) == 0
        if isinstance(default_value, _Param):
            self.default_value = default_value.default_value
        else:
            self.default_value = default_value
        self.lightcurves = lightcurves

        # Check that shape matches up with the given dependencies
        if not self.telescope_dependent and not shape[0] == 1:
            raise ValueError('The ParamArray "{}" is not telescope dependent, but the shape {} implies that it is!'.format(self. name, self.shape))
        if not self.filter_dependent and not shape[1] == 1:
            raise ValueError('The ParamArray "{}" is not filter dependent, but the shape {} implies that it is!'.format(self. name, self.shape))
        if not self.epoch_dependent and not shape[2] == 1:
            raise ValueError('The ParamArray "{}" is not epoch dependent, but the shape {} implies that it is!'.format(self. name, self.shape))

        # Create the array which we will fill with _Params
        if lightcurves is None:
            # No lightcurves provided, we can just fill the array with params
            self.array = np.full(shape, default_value, object)

        else:
            if not lightcurves.shape == shape:
                raise ValueError('shape and lightcurves.shape do not match')

            # Only fill those where the lightcurves are
            self.array = np.full(shape, None, object)
            for i in np.ndindex(lightcurves.shape):
                if lightcurves[i] is not None:
                    self.array[i] = default_value

    def set_value(self, value, telescope_idx=None, filter_idx=None,
                  epoch_idx=None):
        '''
        Sets the array entry at the provided value

        Value can be any object
        '''
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        self.array[idx] = value

    def get_value(self, telescope_idx=None, filter_idx=None,
                  epoch_idx=None):
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        return self.array[idx]

    def add_uniform_fit_param(self, low_lim, high_lim, telescope_idx=None,
                              filter_idx=None, epoch_idx=None,
                              negative_allowed=True):
        '''
        Adds a parameter to be fitted with uniform sampling
        '''
        self.set_value(_UniformParam(low_lim, high_lim, negative_allowed),
                       telescope_idx, filter_idx, epoch_idx)

    def add_gaussian_fit_param(self, mean, stdev, telescope_idx=None,
                               filter_idx=None, epoch_idx=None,
                               negative_allowed=True,
                               clipped_gaussian=False):
        '''
        Adds a gaussian sampled fitting parameter
        '''
        self.set_value(_GaussianParam(mean, stdev, negative_allowed, clipped_gaussian),
                       telescope_idx, filter_idx, epoch_idx)

    def from_unit_interval(self, u, telescope_idx=None, filter_idx=None,
                           epoch_idx=None):
        '''
        Converts a value u in the range [0,1] to a physical value. Used in the
        dynesty routine
        '''
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        return self.array[idx].from_unit_interval(u)

    def clipped_gaussian_transform(self, u, telescope_idx=None, filter_idx=None,
                           epoch_idx=None):
        '''
        Converts a value u in the range [0,1] to a physical value. Used in the
        dynesty routine
        '''
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        return self.array[idx].uniform_to_clipped_gaussian(u)

    def generate_blank_ParamArray(self):
        '''
        Produces a ParamArray with the same telescope, filter, and epoch
        dependencies and shape, but with everything else initialised to None
        '''
        return ParamArray(self.name, self.shape, self.telescope_dependent,
                          self.filter_dependent, self.epoch_dependent)

    def get_flat_array(self):
        '''
        Returns the array as a flat list. Useful for getting average values
        '''
        return self.array.flatten()

    def _generate_idx(self, telescope_idx, filter_idx, epoch_idx):
        '''
        Takes the given indices and converts them into a usable index, omitting
        any that are given for variables that the ParamArray does not depend on
        and raises ValueErrors if None is provided when required.
        '''
        if self.global_param:
            return (0, 0, 0)

        idx = ()
        if self.telescope_dependent:
            if telescope_idx is None:
                raise ValueError('telescope_idx must be provided')
            idx += (telescope_idx,)
        else:
            idx += (0,)

        if self.filter_dependent:
            if filter_idx is None:
                raise ValueError('filter_idx must be provided')
            idx += (filter_idx,)
        else:
            idx += (0,)

        if self.epoch_dependent:
            if epoch_idx is None:
                raise ValueError('epoch_idx must be provided')
            idx += (epoch_idx,)
        else:
            idx += (0,)

        return idx

    def __getitem__(self, idx):
        telescope_idx, filter_idx, epoch_idx = idx
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        return self.array[idx]

    def __setitem__(self, idx, value):
        telescope_idx, filter_idx, epoch_idx = idx
        idx = self._generate_idx(telescope_idx, filter_idx, epoch_idx)
        self.array[idx] = value

    def __str__(self):
        print_str = ''
        for i in np.ndindex(self.shape):
            if self[i] is not None:
                line = '{}:\t'.format(self.name)
                if self.telescope_dependent:
                    line += 't {}\t'.format(i[0])
                else:
                    line += 't -\t'

                if self.filter_dependent:
                    line += 'f {}\t'.format(i[1])
                else:
                    line += 'f -\t'

                if self.epoch_dependent:
                    line += 'e {}\t'.format(i[2])
                else:
                    line += 'e -\t'

                if type(self[i]) is _Param:
                    line += 'Fixed - value: {}\n'.format( self[i].default_value)
                elif type(self[i]) is float or type(self[i]) is int:
                    line += 'Fixed - value: {}\n'.format(self[i])
                elif type(self[i]) is _UniformParam:
                    line += 'Uniform - min: {} - max: {}\n'.format(self[i].low_lim, self[i].high_lim)
                elif type(self[i]) is _GaussianParam:
                    line += 'Gaussian - mean: {} - stdev: {}\n'.format(self[i].mean, self[i].stdev)
                else:
                    line += 'Unrecognised type - {}\n'.format( self[i].__str__())
                print_str += line
        return print_str



'''
Plotting module for TransitFit
'''

import numpy as np
import matplotlib
matplotlib.use('Agg')
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import to_rgb
from matplotlib import colors
import batman
import os

def plot_individual_lightcurves(lightcurves, priorinfo, results,
                                folder_path='./plots', figsize=(12,8),
                                marker_color='dimgrey', line_color='black',
                                titles=None, add_titles=True, fnames=None,
                                plot_phase=False, t0=None, period=None):
    '''
    Once you have run retrieval, use this to plot things!

    Will make a figure for each light curve, and save with filenames reflecting
    telescope, filter, and epoch number

    plot_phase : bool, optional
        If True and t0 and period are provided, will plot using phase rather
        than time on x axis. This is useful for folded curves.
    t0 : float, optional
        The value of t0 that the lightcurves are folded to. This is treated as
        phase 0 (centred in plot).
    period : float, optional
        The period that lightcurves are folded with.
    '''
    # Get some numbers for loop purposes
    n_epochs = priorinfo.n_epochs
    n_filters = priorinfo.n_filters

    if titles is not None:
        try:
            titles = np.array(titles)
            if not titles.shape == lightcurves.shape:
                print('Warning: titles shape is {}, not {}. Reverting to default titles'.format(titles.shape, lightcurves.shape))
                titles = None
        except Exception as e:
            print('Raised exception: {}'.format(e))
            print('Reverting to default titles')
            titles = None


    if fnames is not None:
        try:
            fnames = np.array(fnames)
            if not fnames.shape == (n_filters, n_epochs):
                print('Warning: fnames shape is {}, not ({},{}). Reverting to default fnames'.format(fnames.shape, n_filters, n_epochs))
                fnames = None
        except Exception as e:
            print('Raised exception: {}'.format(e))
            print('Reverting to default fnames')
            fnames = None

    # Get the best values
    best = results.samples[np.argmax(results.logl)]

    best_dict = priorinfo._interpret_param_array(best)

    # Get the array of detrending coeffs:
    if priorinfo.detrend:
        # We need to combine the detrending coeff arrays into one
        # Each entry should be a list containing all the detrending
        # coefficients to trial.
        d = np.full(lightcurves.shape, None, object)

        for i in np.ndindex(d.shape):
            for coeff in np.ravel(priorinfo.detrending_coeffs):
                if best_dict[coeff][i] is not None:
                    if d[i] is None:
                        d[i] = [best_dict[coeff][i]]
                        d_new=[d[i],best_dict['t0'][i], best_dict['P'][i]]
                    else:
                        d[i].append(best_dict[coeff][i])
                        d_new=[*d[i],best_dict['t0'][i], best_dict['P'][i]]

                    

    # For each light curve, make a plot!
    for i in np.ndindex(lightcurves.shape):
        if lightcurves[i] is not None:
            # We are plotting this!!
            telescope_idx = lightcurves[i].telescope_idx
            filter_idx = lightcurves[i].filter_idx
            epoch_idx = lightcurves[i].epoch_idx


            # Set up the figure and the relevant axes
            gs = gridspec.GridSpec(6, 7)
            fig = plt.figure(figsize=figsize)

            main_ax = fig.add_subplot(gs[:-2, :-1])
            residual_ax = fig.add_subplot(gs[-2:, :-1], sharex=main_ax)
            hist_ax = fig.add_subplot(gs[-2:,-1], sharey=residual_ax)

            # Work out x axis values and labels
            if not plot_phase:
                # x axis is time, not phase
                x_vals = lightcurves[i].times
                x_label = 'Time (BJD)'
            else:
                # x axis is time, need to calculate phase
                try:
                    x_vals = (lightcurves[i].times - (t0 + (period/2)))/period
                    x_label = 'Phase'
                except Exception as e:
                    print('Exception raised when calculating phase. Reverting to plotting time')
                    print(e)
                    x_vals = lightcurves[i].times
                    x_label= 'Time (BJD)'

            plot_times = np.linspace(lightcurves[i].times.min(), lightcurves[i].times.max(), 1000 )

            # Now make the best fit light curve

            # First we set up the parameters
            params = batman.TransitParams()
            params.t0 = best_dict['t0'][i]
            params.per = best_dict['P'][i]
            params.rp = best_dict['rp'][i]
            params.a = best_dict['a'][i]
            params.inc = best_dict['inc'][i]
            params.ecc = best_dict['ecc'][i]
            params.w = best_dict['w'][i]
            params.limb_dark = priorinfo.limb_dark

            if priorinfo.fit_ld:
                # NOTE needs converting from q to u
                best_q = np.array([best_dict[key][i] for key in priorinfo.limb_dark_coeffs])
            else:
                q = np.array([priorinfo.priors[key][i] for key in priorinfo.limb_dark_coeffs])
                for j in np.ndindex(q.shape):
                    q[j] = q[j].default_value
                best_q = q

            params.u = priorinfo.ld_handler.convert_qtou(*best_q)

            m = batman.TransitModel(params, plot_times)

            best_curve = m.light_curve(params)

            m_sample_times = batman.TransitModel(params, lightcurves[i].times)
            time_wise_best_curve = m_sample_times.light_curve(params)

            norm = best_dict['norm'][i]

            # Plot the raw data
            if priorinfo.detrend:
                plot_fluxes, plot_errors = lightcurves[i].detrend_flux(d_new, norm)
            else:
                plot_fluxes, plot_errors = lightcurves[i].detrend_flux([None,best_dict['t0'][i], best_dict['P'][i]], norm)

            main_ax.errorbar(x_vals, plot_fluxes,
                        plot_errors, zorder=1,
                        linestyle='', marker='x', color=marker_color,
                        elinewidth=0.8, alpha=0.6)

            # Plot the curve
            # Convert to phase if needed:
            if plot_phase:
                plot_times = (plot_times - t0 + (period/2))/period

            main_ax.plot(plot_times, best_curve, linewidth=2,
                        color=line_color)

            # plot the residuals
            residuals = plot_fluxes - time_wise_best_curve

            residual_ax.errorbar(x_vals, residuals,
                        plot_errors, linestyle='',
                        color=marker_color, marker='x', elinewidth=0.8, alpha=0.6)

            residual_ax.axhline(0, linestyle='dashed', color='gray',
                                linewidth=1, zorder=1)

            # Histogram the residuals
            # Sort out colors:
            rgba_color = colors.to_rgba(marker_color)
            facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)

            hist_ax.hist(residuals, bins=30, orientation='horizontal',
                         color=facecolor, edgecolor=rgba_color,
                         histtype='stepfilled')
            hist_ax.axhline(0, linestyle='dashed', color='gray',
                            linewidth=1, zorder=1)


            # Prune axes
            main_ax.yaxis.set_major_locator(MaxNLocator(6, prune='lower'))
            residual_ax.yaxis.set_major_locator(MaxNLocator(4, prune='upper'))
            residual_ax.xaxis.set_major_locator(MaxNLocator(8, prune='upper'))

            # Add labels
            main_ax.set_ylabel('Normalised flux')
            residual_ax.set_ylabel('Residual')
            residual_ax.set_xlabel(x_label)

            # Format the axes
            main_ax.tick_params('both', which='both', direction='in',
                                labelbottom=False, top=True, right=True)

            residual_ax.tick_params('both', which='both', direction='in',
                                    top=True, right=True)

            hist_ax.tick_params('both', which='both', direction='in',
                                 labelleft=False, labelbottom=False,
                                 right=True, top=True)


            if add_titles:
                if titles is None:
                    if plot_phase:
                        title = 'Phase plot for filter {}'.format(filter_idx)
                    else:
                        title = 'Fitted curve: Telescope {} Filter {}, Epoch {}'.format(telescope_idx, filter_idx, epoch_idx)
                else:
                    title = titles[i]

                main_ax.set_title(title)

            fig.tight_layout()
            fig.subplots_adjust(hspace=0, wspace=0)

            # Save the figures
            # Make the plots folder
            os.makedirs(folder_path, exist_ok=True)

            if fnames is None:
                if plot_phase:
                    fname = 'f{}_phase_plot.pdf'.format(filter_idx)
                else:
                    fname = 't{}_f{}_e{}.pdf'.format(telescope_idx, filter_idx, epoch_idx)
            else:
                if fnames[i] is None:
                    fname = 't{}_f{}_e{}.pdf'.format(telescope_idx, filter_idx, epoch_idx)
                else:
                    if not fnames[i][-4:] == '.pdf':
                        fnames[i] += '.pdf'
                    fname = fnames[i]

            fig.savefig(os.path.join(folder_path, fname), bbox_inches='tight')

            plt.close()


def quick_plot(lightcurve, fname, folder_path, t0=None, period=None):
    '''
    Quickly plots a single lightcurve on some axes

    If t0 and period are provided, will use a phase plot
    '''
    base_fname = fname

    if t0 is None or period is None:
        # Plot the raw times with t0 added
        fig, ax = plt.subplots()
        x_vals = lightcurve.times
        x_label= 'Time (BJD)'
        ax.errorbar(x_vals, lightcurve.flux, lightcurve.errors, zorder=1,
            linestyle='', marker='x', color='dimgrey', elinewidth=0.8, alpha=0.6)

        if t0 is not None:
            ax.axvline(t0, linestyle='dashed', color='gray',
                            linewidth=1, zorder=1)

        ax.set_xlabel(x_label)
        ax.set_ylabel('Normalised flux')

        ax.tick_params('both', which='both', direction='in',
                            labelbottom=True, top='on', right='on')

        os.makedirs(folder_path, exist_ok=True)

        if not base_fname[-4:] == '.png':
            fname = base_fname + '_raw_times.png'
        else:
            fname = base_fname[:-4] + '_raw_times.png'

        fig.savefig(os.path.join(folder_path, fname),
                    bbox_inches='tight')

        return

    # Plot with phases
    # Plot the raw times with t0 added
    fig, ax = plt.subplots()

    x_vals = lightcurve.get_phases(t0, period)
    x_label = 'Phase'

    ax.errorbar(x_vals, lightcurve.flux, lightcurve.errors, zorder=1,
        linestyle='', marker='x', color='dimgrey', elinewidth=0.8, alpha=0.6)

    ax.axvline(0, linestyle='dashed', color='gray',
               linewidth=1, zorder=1)

    ax.set_xlabel(x_label)
    ax.set_ylabel('Normalised flux')

    ax.tick_params('both', which='both', direction='in',
                        labelbottom=True, top='on', right='on')

    os.makedirs(folder_path, exist_ok=True)

    if not base_fname[-4:] == '.png':
        fname = base_fname + '_phase.png'
    else:
        fname = base_fname[:-4] + '_phase.png'

    fig.savefig(os.path.join(folder_path, fname),
                bbox_inches='tight')


def plot_from_file(path, phase_plot=True, folder_path='./plots',
                   figsize=(12,8), marker_color='dimgrey',
                   line_color='black', title=None, save_path=None):
    '''
    Plots a light curve from a file.

    If phase_plot is true, then will plor phase on x axis. Otherwise plots time.

    '''
    # Get the data
    time, phase, flux, flux_err, model = pd.read_csv(path).values.T
    residuals = flux - model

    # Set up the figure and the relevant axes
    gs = gridspec.GridSpec(6, 7)
    fig = plt.figure(figsize=figsize)

    main_ax = fig.add_subplot(gs[:-2, :-1])
    residual_ax = fig.add_subplot(gs[-2:, :-1], sharex=main_ax)
    hist_ax = fig.add_subplot(gs[-2:,-1], sharey=residual_ax)

    if phase_plot:
        x_vals = phase
        x_label = 'Phase'
    else:
        x_vals = time
        x_label = 'Time (BJD)'

    main_ax.errorbar(x_vals, flux,
                     flux_err, zorder=1,
                     linestyle='', marker='x', color=marker_color,
                     elinewidth=0.8, alpha=0.6)

    # Plot the curve
    main_ax.plot(xvals, model, linewidth=2,
                color=line_color)

    # plot the residuals
    residual_ax.errorbar(x_vals, residuals,
                flux_err, linestyle='',
                color=marker_color, marker='x', elinewidth=0.8, alpha=0.6)

    residual_ax.axhline(0, linestyle='dashed', color='gray',
                        linewidth=1, zorder=1)

    # Histogram the residuals
    # Sort out colors:
    rgba_color = colors.to_rgba(marker_color)
    facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)

    hist_ax.hist(residuals, bins=30, orientation='horizontal',
                 color=facecolor, edgecolor=rgba_color,
                 histtype='stepfilled')
    hist_ax.axhline(0, linestyle='dashed', color='gray',
                    linewidth=1, zorder=1)


    # Prune axes
    main_ax.yaxis.set_major_locator(MaxNLocator(6, prune='lower'))
    residual_ax.yaxis.set_major_locator(MaxNLocator(4, prune='upper'))
    residual_ax.xaxis.set_major_locator(MaxNLocator(8, prune='upper'))

    # Add labels
    main_ax.set_ylabel('Normalised flux')
    residual_ax.set_ylabel('Residual')
    residual_ax.set_xlabel(x_label)

    # Format the axes
    main_ax.tick_params('both', which='both', direction='in',
                        labelbottom=False, top=True, right=True)

    residual_ax.tick_params('both', which='both', direction='in',
                            top=True, right=True)

    hist_ax.tick_params('both', which='both', direction='in',
                         labelleft=False, labelbottom=False,
                         right=True, top=True)

    if title is not None:
        main_ax.set_title(title)

    if save_path is None:
        save_path = os.path.splitext(path)[0] + '_plot.pdf'

    fig.savefig(save_path, bbox_inches='tight')

    plt.close()



'''
A class for parameters in TransitFit which can be retrieved. These are used
by the PriorInfo to determine dimensionality etc.

'''
from scipy.special import erf, erfinv
import numpy as np

class _Param:
    def __init__(self, value, uncertainty=None):

        self.default_value = value
        self.low_lim=None
        self.high_lim=None
        self.uncertainty=uncertainty

    def from_unit_interval(self, u):
        raise NotImplementedError


class _UniformParam(_Param):
    def __init__(self, low_lim, high_lim, negative_allowed=True):
        if low_lim >= high_lim:
            raise ValueError('low_lim >= high_lim')

        super().__init__((high_lim + low_lim)/2)
        self.low_lim = low_lim
        self.high_lim = high_lim
        self.negative_allowed = negative_allowed

    def from_unit_interval(self, u):
        '''
        Function to convert value u in range (0,1], will convert to a value to
        be used by Batman
        '''
        if u > 1 or u < 0:
            raise ValueError('u must satisfy 0 < u < 1. ')
        val = u * (self.high_lim - self.low_lim) + self.low_lim
        if self.negative_allowed:
            return val
        return abs(val)

class _GaussianParam(_Param):
    def __init__(self, best, sigma, negative_allowed=True, clipped_gaussian=False):
        '''
        A GaussianParam is one which is fitted using a Gaussian prior (normal)
        distribution.
        '''

        super().__init__(best)
        self.mean = best
        self.stdev = sigma
        self.negative_allowed = negative_allowed

        if clipped_gaussian:
            self.factor_for_clipping=(erf((90 - self.mean)/(self.stdev * np.sqrt(2))) + 1)/2 
        

    def from_unit_interval(self, u):
        '''
        Function to convert value u in range (0,1], will convert to a value to
        be used by Batman
        '''
        if u > 1 or u < 0:
            raise ValueError('u must satisfy 0 < u < 1')

        val = self.mean + self.stdev * np.sqrt(2) * erfinv(2 * u - 1)
        if self.negative_allowed:
            return val
        return abs(val)
    
    def uniform_to_clipped_gaussian(self,u):
        '''
        Function to convert value u in range (0,1], will convert to a value to
        be used by Batman
        '''
        if u > 1 or u < 0:
            raise ValueError('u must satisfy 0 < u < 1')

        val = self.mean + self.stdev * np.sqrt(2) * erfinv(2 * u * self.factor_for_clipping - 1)
        if self.negative_allowed:
            return val
        return abs(val)



'''
Object to deal with writing fitting results to files
'''
import numpy as np
import os
import csv
import pandas as pd
import batman
import itertools
import traceback
import corner
import pickle
from collections.abc import Iterable


import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import to_rgb
from matplotlib import colors

from .retriever import global_params, filter_dependent_params, lightcurve_dependent_params
from ._utils import weighted_avg_and_std, host_radii_to_AU, get_normalised_weights, get_covariance_matrix
from ._paramarray import ParamArray
from .lightcurve import LightCurve
from .error_analysis import ErrorLimits, get_quantiles_on_best_val


class OutputHandler:
    '''
    Designed to handle writing outputs from retrieval to files

    Parameters
    ----------
    lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
        The full lightcurves array to be retrieved.
    full_prior : PriorInfo
        The prior for the complete light curve dataset.

    '''
    def __init__(self, lightcurves, full_prior, host_r=None):

        self.all_lightcurves = lightcurves

        self.full_prior = full_prior

        self.ld_coeffs = self.full_prior.limb_dark_coeffs
        self.ld_coeffs_u = ['u{}'.format(i) for i in range(len(self.ld_coeffs))]
        self.ld_model = self.full_prior.limb_dark

        self.host_r = host_r

        self.best_model = None
        self.batman_initialised = False

        self.global_params = []
        for i, param in enumerate(self.full_prior.fitting_params):
            if np.all(param[1:] == None):
                self.global_params.append(param[0])

    def save_final_light_curves(self, all_lightcurves, global_prior,
                                folder='./final_light_curves', folded=False):
        '''
        Saves the final curves and best fit model to file

        Requires best model to be initialised

        Parameters
        ----------
        all_lightcurves : array_like
            Array containing the global set of light curves - these should be
            raw and not normalised or detrended.
        '''
        print('Saving final light curves')
        # Set up the model
        self._initialise_batman(all_lightcurves)

        # Put all the detrending coeffs in usable format
        d = np.full(all_lightcurves.shape, None, object)

        if global_prior.detrend:
            # We need to combine the detrending coeff arrays into one
            # Each entry should be a list containing all the detrending
            # coefficients for the light curve.
            for i in np.ndindex(d.shape):
                det_coeff_flattened = sum([det for det in global_prior.detrending_coeffs], [])
                for coeff in det_coeff_flattened: #np.ravel(global_prior.detrending_coeffs):                   
                    if type(coeff) is not list:
                        if self.best_model[coeff][i] is not None:
                            if d[i] is None:
                                d[i] = [self.best_model[coeff][i][0]]

                            else:
                                d[i].append(self.best_model[coeff][i][0])
                    else:
                        for coeff_i in coeff:
                            if self.best_model[coeff_i][i] is not None:
                                if d[i] is None:
                                    d[i] = [self.best_model[coeff_i][i][0]]

                                else:
                                    d[i].append(self.best_model[coeff_i][i][0])
                            

        for i, lc in np.ndenumerate(all_lightcurves):
            # Loop through each light curve, make the best model, and save it!
            if lc is not None:
                # First, detrend and normalise the curve

                try:    
                    d_new=[*d[i],self.best_model['t0'][i][0], self.best_model['P'][i][0]]
                except TypeError:
                    d_new=[d[i],self.best_model['t0'][i][0], self.best_model['P'][i][0]]
                #flux, flux_err = lc.detrend_flux(d[i], self.best_model['norm'][i][0], force_normalise=True)
                flux, flux_err = lc.detrend_flux(d_new, self.best_model['norm'][i][0], force_normalise=True)

                # Get phase
                phase = lc.get_phases(self.best_model['t0'][i][0], self.best_model['P'][i][0])

                # Get the best fit model depths
                model_curve = self.batman_models[i].light_curve(self.batman_params[i])

                write_data = np.vstack((lc.times, phase, flux, flux_err, model_curve)).T

                write_data_frame = pd.DataFrame(write_data, columns=['Time', 'Phase', 'Normalised flux', 'Flux uncertainty', 'Best fit curve'])

                save_path = os.path.join(folder,'t{}_f{}_e{}_detrended.csv'.format(*i))

                os.makedirs(folder, exist_ok=True)
                write_data_frame.to_csv(save_path, index=False, na_rep='-')

    def plot_final_light_curves(self, all_lightcurves, global_prior,
                                folder='./plots', figsize=(12,8),
                                marker_color='dimgrey', line_color='black',
                                plot_folded=True, titles=False, bin_data=True,
                                cadence=2, binned_color='red'):
        '''
        Plots the detrended light curves with the global best-fit model

        Parameters
        ----------
        cadence : float, optional
            The cadence to bin to in minutes

        '''
        print('Plotting final curves')

        # Getting the maximum limits of the phase in all light curves
        phase_lims=[1,-1]
        for i, lc in np.ndenumerate(all_lightcurves):
            # Loop through each light curve, make the best model, and save it!
            if lc is not None:
                # Get phase
                phase = lc.get_phases(self.best_model['t0'][i][0], self.best_model['P'][i][0])
                phase_lims[0]=min(phase_lims[0],min(phase))
                phase_lims[1]=max(phase_lims[1],max(phase))
        phase_lims = [phase_lims[0]-0.005, phase_lims[1]+0.005] # Changing the limits by 0.005 to have a small gap

        # First, deal with detrending
        # Put all the detrending coeffs in usable format
        d = np.full(all_lightcurves.shape, None, object)

        if global_prior.detrend:
            # We need to combine the detrending coeff arrays into one
            # Each entry should be a list containing all the detrending
            # coefficients for the light curve.
            for i in np.ndindex(d.shape):
                det_coeff_flattened = sum([det for det in global_prior.detrending_coeffs], [])
                for coeff in det_coeff_flattened: #np.ravel(global_prior.detrending_coeffs):
                    if type(coeff) is not list:
                        if self.best_model[coeff][i] is not None:
                            if d[i] is None:
                                d[i] = [self.best_model[coeff][i][0]]

                            else:
                                d[i].append(self.best_model[coeff][i][0])
                    else:
                        for coeff_i in coeff:
                            if self.best_model[coeff_i][i] is not None:
                                if d[i] is None:
                                    d[i] = [self.best_model[coeff_i][i][0]]

                                else:
                                    d[i].append(self.best_model[coeff_i][i][0])


        # Now we loop through each curve and make the plots
        for i, lc in np.ndenumerate(all_lightcurves):
            # Loop through each light curve, make the best model, and save it!
            if lc is not None:
                # First, detrend and normalise the curve
                try:    
                    d_new=[*d[i],self.best_model['t0'][i][0], self.best_model['P'][i][0]]
                except TypeError:
                    d_new=[d[i],self.best_model['t0'][i][0], self.best_model['P'][i][0]]
                #flux, flux_err = lc.detrend_flux(d[i], self.best_model['norm'][i][0], force_normalise=True)
                flux, flux_err = lc.detrend_flux(d_new, self.best_model['norm'][i][0], force_normalise=True)

                # Get phase
                phase = lc.get_phases(self.best_model['t0'][i][0], self.best_model['P'][i][0])

                # Get the best fit model depths - use linspaced times for plot
                # and the lc times for residuals
                model_times = np.linspace(lc.times.min(), lc.times.max(), 1000)
                model = batman.TransitModel(self.batman_params[i], model_times)
                model_curve = model.light_curve(self.batman_params[i])
                time_wise_best_curve = self.batman_models[i].light_curve(self.batman_params[i])
                # get model phase:
                n = (model_times - (self.best_model['t0'][i][0] - 0.5 * self.best_model['P'][i][0]))//self.best_model['P'][i][0]

                model_phase = (model_times - self.best_model['t0'][i][0])/self.best_model['P'][i][0] - n# + 0.5

                # Get the residual
                residuals = flux - time_wise_best_curve

                plot_errors = [flux_err, None]
                sub_folder = ['with_errorbars', 'without_errorbars']

                for j in range(2):
                    fname = 'individual_curves/{}/t{}_f{}_e{}.png'.format(sub_folder[j],*i)
                    if titles:
                        title = 'Fitted curve: Telescope {} Filter {}, Epoch {}'.format(*i)
                    else:
                        title = None

                    ####### PLOT! #########
                    self._plot_data(phase, flux, plot_errors[j], model_phase,
                                    model_curve, residuals, fname, title, folder,
                                    figsize, marker_color, line_color, phase_lims)

        if not plot_folded:
            return

        # Now we fold the light curves for each filter
        for fi in range(all_lightcurves.shape[1]):
            phase = []
            flux = []
            flux_err = []
            residuals = []
            model_phase = []
            model_flux = []

            for ti, ei in itertools.product(range(all_lightcurves.shape[0]), range(all_lightcurves.shape[2])):
                i = (ti, fi, ei)
                lc = all_lightcurves[i]
                if lc is not None:
                    try:    
                        d_new=[*d[i],self.best_model['t0'][i][0],self.best_model['P'][i][0]]
                    except TypeError:
                        d_new=[d[i],self.best_model['t0'][i][0],self.best_model['P'][i][0]]

                    #lc_flux, lc_flux_err = lc.detrend_flux(d[i], self.best_model['norm'][i][0], force_normalise=True)
                    lc_flux, lc_flux_err = lc.detrend_flux(d_new, self.best_model['norm'][i][0], force_normalise=True)
                    lc_phase = lc.get_phases(self.best_model['t0'][i][0], self.best_model['P'][i][0])

                    # Get the best fit model depths - use linspaced times for plot
                    # and the lc times for residuals
                    model_times = np.linspace(lc.times.min(), lc.times.max(), 1000)
                    model = batman.TransitModel(self.batman_params[i], model_times)
                    lc_model_curve = model.light_curve(self.batman_params[i])
                    time_wise_best_curve = self.batman_models[i].light_curve(self.batman_params[i])

                    # get model phase:
                    n = (model_times - (self.best_model['t0'][i][0] - 0.5 * self.best_model['P'][i][0]))//self.best_model['P'][i][0]

                    lc_model_phase = (model_times - self.best_model['t0'][i][0])/self.best_model['P'][i][0] - n# + 0.5


                    lc_residuals = lc_flux - time_wise_best_curve

                    # Store the values!
                    phase += list(lc_phase)
                    flux += list(lc_flux)
                    flux_err += list(lc_flux_err)
                    residuals += list(lc_residuals)
                    model_phase += list(lc_model_phase)
                    model_flux += list(lc_model_curve)

            phase = np.array(phase).flatten()
            flux = np.array(flux).flatten()
            flux_err = np.array(flux_err).flatten()
            residuals = np.array(residuals).flatten()
            model_phase = np.array(model_phase).flatten()
            model_flux = np.array(model_flux).flatten()

            P = self.best_model['P'][None, None, None][0]
            t0 = self.best_model['t0'][None, None, 0][0]

            cadence_days = cadence / (24 * 60)

            cadence_phase = cadence_days/P

            plot_errors = [flux_err, None]
            sub_folder = ['with_errorbars', 'without_errorbars']

            for j in range(2):
                fname = 'folded_curves/{}/filter_{}.png'.format(sub_folder[j],fi)
                if titles:
                    title = 'Folded curve for filter {}'.format(fi)
                else:
                    title = None

                ####### PLOT! #########
                try:
                    self._plot_data(phase, flux, plot_errors[j], model_phase,
                                model_flux, residuals, fname, title, folder,
                                    figsize, marker_color, line_color, phase_lims, bin_data,
                                cadence_phase, binned_color)

                except Exception as e:
                    print('Exception raised while plotting:')
                    print(e)
                    print('Traceback:')
                    traceback.print_tb(e.__traceback__)

    def save_complete_results(self, mode, global_prior, output_folder,
                              summary_file):
        '''
        Once all batches etc are run, collates all results and saves to csv
        '''
        _ = self._initialise_best_model(mode, global_prior, output_folder, summary_file)

        print('Saving final results')

        self._save_results_dict(self.best_model, os.path.join(output_folder, 'Complete_results.csv'), False)

        try:
            el = ErrorLimits(output_folder)
            el.get_errors()
        except:
            print("An exception occurred while trying to get asymmetric errors!")


    def save_results(self, results, priors, lightcurves,
                     output_folder='./output_parameters',
                     summary_file='summary_output.csv',
                     full_output_file='full_output.csv',
                     samples_plot_folder='./plots', folded=False,
                     plot_posteriors=True, batch_idx=None, stage=1):
        '''
        Saves results to .csv files

        Parameters
        ----------
        results : array_like, shape (n_batches, )
            The results from each run
        priors : array_like, shape (n_batches, )
            The priors for each run
        lightcurves : array_like, shape (n_batches, )
            The light curves for each run
        fit_ld : bool, optional
            Must be true if LDCs are fitted. Default is True
        output_folder : str, optional
            The folder to save output files to (not plots). Default is
            `'./output_parameters'`
        summary_file : str, optional
            The file name for the final output. This file only gives the
            averaged values, rather than individual values fitted within
            batches if there are any. Default is `'summary_output.csv'`
        full_output_file : str, optional
            The file name for the full output file. This file gives partial
            results from batches, rather than the averaged results. Default is
            `'full_output.csv'`

        Returns
        -------
        best_vals : dict
            Each entry is [best val, error]
        combined_dict : dict
            The combined results dictionary. Each entry is a list of values
            from the results dictionaries -
            [[best value, median, 16th percentile, 84th percentile, stdev],...]
        '''
        print('Saving full results...')
        fit_ld = priors[0].fit_ld

        results_dicts = []

        for i, ri in enumerate(results):
            results_dicts.append(self.get_results_dict(ri, priors[i], lightcurves[i]))
            if plot_posteriors:
                try:
                    if folded:
                        sample_folder = os.path.join(samples_plot_folder, 'folded')
                    else:
                        sample_folder = os.path.join(samples_plot_folder, 'unfolded')
                    print(f'Plotting batch {i} samples to {os.path.join(sample_folder, f"batch_{i}_samples.png")}')
                    self._plot_samples(ri, priors[i], f'batch_{i}_samples.png', sample_folder)
                except Exception as e:
                    print(e)
        best_vals, combined_results = self.get_best_vals(results_dicts, fit_ld)

        print(f'Saving summary results to {os.path.join(output_folder, summary_file)}')
        self._save_results_dict(best_vals, os.path.join(output_folder, summary_file), False)
        print(f'Saving full results to {os.path.join(output_folder, full_output_file)}')
        self._save_results_dict(combined_results, os.path.join(output_folder, full_output_file), True)
        print('Results saved')

        return best_vals, combined_results

    def get_results_dict(self, results, priors, lightcurves):
        '''
        Makes dictionaries of results from a single dynesty run.

        Useful for putting results in a form which allows for easy summary etc.

        Parameters
        ----------
        results : dict
            Results from a single dynesty run
        priors : PriorInfo
            The priors for the run
        lightcurves : array_like, shape (n_telescopes, n_filters, n_epochs)
            The lightcurves for the run

        Returns
        -------
        results_dict : dict
            Each entry is [best value, median, 16th percentile, 84th percentile, stdev]
        '''
        print('Extracting results dict')
        results_dict = {}

        # Loop over all fitting parameters and access the results
        for i, param_info in enumerate(priors.fitting_params):
            param_name, batch_tidx, batch_fidx, batch_eidx = param_info

            batch_idx = (batch_tidx, batch_fidx, batch_eidx)
            # GET INDICES
            # The indices here are for a particular batch. We want global
            # values so pull them out of the LightCurves
            full_idx = self._batch_to_full_idx(batch_idx, param_name, lightcurves, priors.allow_ttv)

            result_entry = [results.best[i], results.median[i], results.lower_err[i], results.upper_err[i], results.uncertainties[i]]

            # Check that the parameter has been initialised in the dict
            results_dict = self._initialise_dict_entry(results_dict, param_name)
            if param_name=='norm':
                self.full_idx_escale=full_idx
            #breakpoint()
            #if param_name=='escale':
            #    full_idx=self.full_idx_escale
            #print(results_dict,param_name,full_idx)
            if results_dict[param_name][full_idx] is None:
                # Initialise a list
                results_dict[param_name][full_idx] = []

            results_dict[param_name][full_idx].append(result_entry)

        # Go through the PriorInfo and get the constant values out
        # These are ones that were in the priors and NOT in fitting params
        for param_name in priors.priors:
            if param_name not in priors.fitting_params:
                # This is a constant value
                # Check that the parameter has been initialised in the dict
                results_dict = self._initialise_dict_entry(results_dict, param_name)
                for i in np.ndindex(priors.priors[param_name].shape):
                    if priors.priors[param_name][i] is not None:
                        result_entry = [priors.priors[param_name].default_value, None, None, None, None]

                        if results_dict[param_name][i] is None:
                            # Initialise a list
                            results_dict[param_name][i] = []

                        results_dict[param_name][i].append(result_entry)
        return results_dict

    def get_best_vals(self, results_dicts, priors, fit_ld=True, return_combined=True):
        '''
        Gets the best values for a set of runs from the given results dicts

        Parameters
        ----------
        results_dicts : array_like, shape (n_batches, )
            The results_dicts obtained from get_results_dict
        fit_ld : bool, optional
            Should be True if LDC fitting. Default is True
        return_combined : bool, optional
            If True, will return the results dicts combined into a single dict.
            Default is True.

        Returns
        -------
        best_vals : dict
            Each entry is [best val, error]
        combined_dict : dict
            The combined results dictionary. Returned if return_combined is
            True
        '''
        print('Calculating best values for this run')
        best_vals = {}

        # Collate the results dicts
        combined_dict = self.combine_results_dicts(results_dicts)

        for param in combined_dict:
            # Loop through each parameter
            best_vals = self._initialise_dict_entry(best_vals, param)

            for i in np.ndindex(combined_dict[param].shape):
                if combined_dict[param][i] is not None:
                    # get the weighted avrage and error
                    if np.any(combined_dict[param][i][:,-1] == None):
                        # This is a constant
                        best_vals[param][i] = (combined_dict[param][i][:,0][0], None)
                    else:
                        best_vals[param][i] = weighted_avg_and_std(combined_dict[param][i][:, 0], combined_dict[param][i][:, -1], single_val=True)

        # Limb darkening bits
        if fit_ld:
            best_vals, combined_dict = self.add_best_u(best_vals, combined_dict)
        if return_combined:
            return best_vals, combined_dict
        return best_vals

    def combine_results_dicts(self, results_dicts):
        '''
        Combines the given results dicts into one dict

        Parameters
        ----------
        results_dicts : array_like, shape (n_batches, )
            The results_dicts obtained from get_results_dict

        Returns
        -------
        combined_dict : dict
            The combined results dictionary. Each entry is a list of values
            from the results dictionaries -
            [[best value, median, 16th percentile, 84th percentile, stdev],...]
        '''
        print('Combining results dicts')
        combined_dict = {}
        # Loop through each dict and the params
        for rd in results_dicts:
            for param in rd.keys():
                combined_dict = self._initialise_dict_entry(combined_dict, param)

                for i in np.ndindex(combined_dict[param].shape):
                    if rd[param][i] is not None:
                        if combined_dict[param][i] is None:
                            combined_dict[param][i] = rd[param][i]
                        else:
                            combined_dict[param][i] += rd[param][i]

        # Convert to np.arrays
        for param in combined_dict.keys():
            for i in np.ndindex(combined_dict[param].shape):
                if combined_dict[param][i] is not None:
                    combined_dict[param][i] = np.array(combined_dict[param][i])

        return combined_dict

    def add_best_u(self, best_dict, combined_dict):
        '''
        Given results dicts, adds in the u vals

        Parameters
        ----------
        best_dict : dict
            The dictionary of best results
        combined_dict : dict
            The full results

        Returns
        -------
        best_dict : dict
            The same dictionary with the best u vals added.
        combined_dict : dict
            The same dictionary with the u vals added.
        '''
        # Initialise each of the u params
        u_coeffs = []
        for param in self.ld_coeffs:
            ldc_q = 'q{}'.format(param[-1])
            ldc_u = 'u{}'.format(param[-1])
            u_coeffs.append(ldc_u)

            if ldc_u not in best_dict:
                combined_dict[ldc_u] = combined_dict[ldc_q].generate_blank_ParamArray()
                best_dict[ldc_u] = combined_dict[ldc_u].generate_blank_ParamArray()

        # Now get the values!
        for i in np.ndindex(best_dict['q0'].shape):
            if best_dict['q0'][i] is not None:
                # There are q values associated with this filter
                for u in u_coeffs:
                    best_dict[u][i] = []
                    combined_dict[u][i] = []

                # Put all the q values for a given filter into one place so we
                # can access q0, q1 simultaneously for converting to u
                n_batches = len(combined_dict['q0'][i])

                # indexing is filter_q[batch, val/err, qX]
                filter_q = np.zeros((n_batches, 2, len(self.ld_coeffs)))

                for b in range(n_batches):
                    for qi, q in enumerate(self.ld_coeffs):
                        filter_q[b,0,qi] = combined_dict[q][i][b][0]
                        filter_q[b,1,qi] = combined_dict[q][i][b][-1]

                # indexing is best_filter_q[val/err, qX]
                best_filter_q = np.vstack([best_dict[q][i] for q in self.ld_coeffs]).T

                # Convert the q values. First up the combined dict:
                for b in range(n_batches):
                    u, u_err = self.full_prior.ld_handler.convert_qtou_with_errors(*filter_q[b])

                    for k, uk in enumerate(u_coeffs):
                        combined_dict[uk][i].append([u[k], u_err[k]])

                # Now the best dict:
                u, u_err = self.full_prior.ld_handler.convert_qtou_with_errors(*best_filter_q)
                for k, uk in enumerate(u_coeffs):
                    best_dict[uk][i] = np.array([u[k], u_err[k]])

        return best_dict, combined_dict

    def _initialise_best_model(self, mode, global_prior, output_folder,
                               summary_file):
        '''
        Sets up the best fit model used to output best fit

        Parameters
        ----------
        mode : str
            The fitting mode being used
        global_prior : PriorInfo
            The full_prior from the retriever. Used to determine array shapes.

        Returns
        -------
        best_model : dict
            A dictionary containing param arrays of the best values

        Notes
        -----
        To do this, we go through all output parameter files and pull in the
        values. In folded mode, we pull in from the global summary first and
        add from the filter summaries after
        '''
        print('Initialising best fit model')
        mode = mode.lower()
        if mode not in ['all', 'batched', 'folded', '2_stage']:
            raise ValueError('Unrecognised mode {}'.format(mode))

        best_model_dict = {}

        # First we initialise each entry in the dict
        for param in global_prior.priors.keys():
            # Initialise from the global prior
            best_model_dict = self._initialise_dict_entry(best_model_dict, param, global_prior)

        for param in self.ld_coeffs:
            # Now initialise the LDC u params
            ldc_q = 'q{}'.format(param[-1])
            ldc_u = 'u{}'.format(param[-1])

            if ldc_u not in best_model_dict:
                best_model_dict[ldc_u] = best_model_dict[ldc_q].generate_blank_ParamArray()

        # Now we go through each of the output files and use them to populate
        # the best_model_dict

        # First use the top-level output
        top_output = pd.read_csv(os.path.join(output_folder, summary_file))

        # List of parameters which were fitted from the top level
        top_params = []

        for i, row in top_output.iterrows():
            param, tidx, fidx, eidx, best, err = row

            if param[-3:] == '/r*':
                param = param[:-3]
            if not err == '-' or mode in ['all', 'batched']:
                if tidx == '-':
                    tidx == None
                else:
                    tidx = int(tidx)

                if fidx == '-':
                    fidx == None
                else:
                    fidx = int(fidx)

                if eidx == '-':
                    eidx == None
                else:
                    eidx = int(eidx)

                best = float(best)
                if err == '-':
                    err = None
                else:
                    err = float(err)

                if param in best_model_dict:
                    if best_model_dict[param][tidx, fidx, eidx] is None:
                        best_model_dict[param][tidx, fidx, eidx] = []

                    best_model_dict[param][tidx, fidx, eidx].append([best, err])

                    if param not in top_params:
                        top_params.append(param)
                #if param in best_model_dict:
                #    best_model_dict[param][tidx, fidx, eidx] = [best, err]

        #if mode in ['all', 'batched']:
        #    self.best_model = best_model_dict
        #    return best_model_dict

        if mode == 'folded':
            # Now we have to go through the results for each of the filters and
            # add in the results from those
            for fi in range(global_prior.n_filters):
                path = os.path.join(output_folder, 'filter_{}_parameters'.format(fi), 'filter_{}_summary.csv'.format(fi))

                filter_output = pd.read_csv(path)
                for i, row in filter_output.iterrows():
                    param, tidx, fidx, eidx, best, err = row

                    if param[-3:] == '/r*':
                        param = param[:-3]

                    if tidx == '-':
                        tidx == None
                    else:
                        tidx = int(tidx)

                    if fidx == '-':
                        fidx == None
                    else:
                        fidx = int(fidx)

                    if eidx == '-':
                        eidx == None
                    else:
                        eidx = int(eidx)

                    best = float(best)
                    if err == '-':
                        err = None
                    else:
                        err = float(err)

                    if param in best_model_dict and param not in top_params:
                        # Store the value(s) from each if the parameter was not
                        # fitted in the folded run
                        if best_model_dict[param][tidx, fidx, eidx] is None:
                            best_model_dict[param][tidx, fidx, eidx] = []

                        best_model_dict[param][tidx, fidx, eidx].append([best, err])


        # Now we go through the params, checking to see if there are multiple
        # values. If there are, we need to take the weighted final values
        for param in best_model_dict:
            for i in np.ndindex(best_model_dict[param].shape):
                if best_model_dict[param][i] is not None:
                    param_results = np.array(best_model_dict[param][i])

                    if param_results[0,1] is None:
                        # deal with the global fixed values (err=None)
                        best_model_dict[param][i] = [param_results[0,0], None]
                        break
                    else:
                        best_model_dict[param][i] = weighted_avg_and_std(param_results[:, 0], param_results[:,1], single_val=True)

        self.best_model = best_model_dict

        return best_model_dict

    def _quicksave_result(self, results, priors, lightcurves,
                          base_output_path='./outputs', filter=None, batch=None):
        '''
        Quickly saves a batch result to file and pickle the Result and Prior
        objects
        '''
        result_dict = self.get_results_dict(results, priors, lightcurves)
        result_dict, _ = self.get_best_vals([result_dict], priors.fit_ld)
        base_fname = ''
        if filter is not None:
            base_fname += f'filter_{filter}_'
        else:
            base_fname += 'all_filters_'
        if batch is not None:
            base_fname += f'batch_{batch}_'
        result_file_fname = base_fname + 'output.csv'
        result_pickle_fname = base_fname +'results.pkl'
        priors_pickle_fname = base_fname + 'priors.pkl'

        # Quicksave best results
        output_path = os.path.join(base_output_path, 'quicksaves', result_file_fname)
        print(f'Quicksaving best results to {output_path}')
        self._save_results_dict(result_dict, output_path, False)

        # Quicksave full results object
        output_path = os.path.join(base_output_path, 'quicksaves', result_pickle_fname)
        print(f'Quicksaving full results to {output_path}')
        with open(output_path, 'wb') as f:
            try:
                pickle.dump(results, f, pickle.HIGHEST_PROTOCOL)
            except Exception as e:
                print('Exception encountered:', e)

        # Quicksave priors
        output_path = os.path.join(base_output_path, 'quicksaves', priors_pickle_fname)
        print(f'Quicksaving priors to {output_path}')
        with open(output_path, 'wb') as f:
            try:
                pickle.dump(priors, f, pickle.HIGHEST_PROTOCOL)
            except Exception as e:
                print('Exception encountered:', e)

    def _save_results_dict(self, results_dict, path, batched):
        '''
        Saves a dict to csv
        '''
        df = self._results_dict_to_dataframe(results_dict, batched)

        # Save outputs
        os.makedirs(os.path.dirname(path), exist_ok=True)
        df.to_csv(path, index=False, na_rep='-')

    def _initialise_batman(self, all_lightcurves):
        '''
        Sets up batman so we can generate best-fit models for outputs

        Parameters
        ----------
        all_lightcurves : array_like
            Array containing the global set of light curves
        '''
        if self.best_model is None:
            raise ValueError('best-fit model is not intiialised.')

        if self.batman_initialised:
            return

        # Check that best model initialisation worked
        failed_key = []
        failed_index = []
        for key in self.best_model.keys():
            for i, lc in np.ndenumerate(all_lightcurves):
                if self.best_model[key][i] is None and lc is not None:
                    # If the failed key is a detrending coeff, we need to do
                    # some extra checking:
                    if key[0] == 'd':
                        method_idx = self.full_prior._detrend_method_index_array[i]
                        if key in self.full_prior.detrending_coeffs[method_idx]:
                            # If the key is associated with the lc, then this
                            # has failed.
                            failed_key.append(key)
                            failed_index.append(i)
                    #elif key[0] == 'norm':
                    else:
                        failed_key.append(key)
                        failed_index.append(i)
        if len(failed_key) > 0:
            print('Best model failed keys:', failed_key)
            print('Best model failed indices:', failed_index)
            raise RuntimeError('Something has gone wrong with the best model generation')


        print('Initialising best batman models')

        n_telescopes = all_lightcurves.shape[0]
        n_filters = all_lightcurves.shape[1]
        n_epochs = all_lightcurves.shape[2]

        # Set up a param array with the best values in
        self.batman_params = ParamArray('batman_params', (n_telescopes, n_filters, n_epochs), True, True, True, lightcurves=all_lightcurves)
        self.batman_models = ParamArray('batman_models', (n_telescopes, n_filters, n_epochs), True, True, True, lightcurves=all_lightcurves)

        for i in np.ndindex(all_lightcurves.shape):
            if all_lightcurves[i] is not None:

                # Set up the parameters
                transit_params = batman.TransitParams()
                transit_params.t0 = self.best_model['t0'][i][0]
                transit_params.per = self.best_model['P'][i][0]
                transit_params.rp = self.best_model['rp'][i][0]
                transit_params.a = self.best_model['a'][i][0]
                transit_params.inc = self.best_model['inc'][i][0]
                transit_params.ecc = self.best_model['ecc'][i][0]
                transit_params.w = self.best_model['w'][i][0]
                transit_params.u = [self.best_model[uX][i][0] for uX in self.ld_coeffs_u]
                transit_params.limb_dark = self.ld_model

                # Save the parameters
                self.batman_params[i] = transit_params

                # Make the model
                self.batman_models[i] = batman.TransitModel(transit_params, all_lightcurves[i].times)

        self.batman_initialised = True

    def _initialise_dict_entry(self, d, param, prior=None):
        '''
        Initialises param in results dictionaries using ParamArray

        If prior is not provided, defaults to self.full_prior
        '''
        if prior is None:
            prior = self.full_prior

        if param not in d:
            d[param] = prior.priors[param].generate_blank_ParamArray()
        return d

    def _batch_to_full_idx(self, i, param_name, lightcurves, allow_ttv):
        '''
        Converts a batch index into a full index

        Parameter
        ---------
        i : tuple
            (batch_tidx, batch_fidx, batch_eidx)

        Returns
        -------
        (tidx, fidx, eidx)
        '''
        batch_tidx, batch_fidx, batch_eidx = i

        if batch_tidx is None:
            tidx = None
        else:
            for k in np.ndindex(lightcurves.shape):
                if k[0] == batch_tidx and lightcurves[k] is not None:
                    tidx = lightcurves[k].telescope_idx

        if batch_fidx is None:
            fidx = None
        else:
            for k in np.ndindex(lightcurves.shape):
                if k[1] == batch_fidx and lightcurves[k] is not None:
                    fidx = lightcurves[k].filter_idx

        if batch_eidx is None:
            eidx = None
        else:
            for k in np.ndindex(lightcurves.shape):
                if k[2] == batch_eidx and lightcurves[k] is not None:
                    eidx = lightcurves[k].epoch_idx

        return tidx, fidx, eidx

    def _results_dict_to_dataframe(self, results_dict, batched):
        '''
        Converts a results dict to a pandas dataframe
        '''
        if batched:
            vals_arr = np.zeros((1, 7), dtype=object)
            columns = ['Parameter', 'Telescope', 'Filter', 'Epoch', 'Batch', 'Best', 'Error']
        else:
            vals_arr = np.zeros((1, 6), dtype=object)
            columns = ['Parameter', 'Telescope', 'Filter', 'Epoch', 'Best', 'Error']

        for param in results_dict:
            # Sort out display of the parameters
            if param == 'rp':
                print_param = 'rp/r*'
            elif param == 'a':
                print_param = 'a/r*'
            else:
                print_param = param

            for i in np.ndindex(results_dict[param].shape):

                if results_dict[param][i] is not None:

                    tidx, fidx, eidx = None, None, None
                    if results_dict[param].telescope_dependent:
                        tidx = i[0]
                    if results_dict[param].filter_dependent:
                        fidx = i[1]
                    if results_dict[param].epoch_dependent:
                        eidx = i[2]

                    if not batched:
                        # Add the best values in
                        if isinstance(results_dict[param][i][0], Iterable):
                            # DIRTY HACK due to weird behaviour with
                            # results from 'all' mode. I'm sorry, but I'm
                            # writing a thesis and don't have the time to
                            # find the actual cause.
                            vals_arr = np.append(vals_arr, np.array([[print_param, tidx, fidx, eidx, results_dict[param][i][0][0], results_dict[param][i][0][-1]]]), axis=0)
                        else:
                            vals_arr = np.append(vals_arr, np.array([[print_param, tidx, fidx, eidx, results_dict[param][i][0], results_dict[param][i][-1]]]), axis=0)

                        if param == 'a' and self.host_r is not None:
                            # Put a into AU as well
                            if isinstance(results_dict[param][i][0], Iterable):
                                # DIRTY HACK due to weird behaviour with
                                # results from 'all' mode
                                a_AU, a_AU_err = host_radii_to_AU(results_dict[param][i][0][0],
                                                                  self.host_r[0],
                                                                  results_dict[param][i][0][1],
                                                                  self.host_r[1], True)
                            else:
                                a_AU, a_AU_err = host_radii_to_AU(results_dict[param][i][0],
                                                                  self.host_r[0],
                                                                  results_dict[param][i][1],
                                                                  self.host_r[1], True)
                            vals_arr = np.append(vals_arr, np.array([['a/AU', tidx, fidx, eidx, a_AU, a_AU_err]]), axis=0)
                    else:
                        # Loop over batches
                        for bi in range(len(results_dict[param][i])):
                            vals_arr = np.append(vals_arr, np.array([[print_param, tidx, fidx, eidx, bi, results_dict[param][i][bi][0], results_dict[param][i][bi][-1]]]), axis=0)

                            if param == 'a' and self.host_r is not None:
                                # Put a into AU as well
                                a_AU, a_AU_err = host_radii_to_AU(results_dict[param][i][bi][0],
                                                                  self.host_r[0],
                                                                  results_dict[param][i][bi][1],
                                                                  self.host_r[1], True)
                                vals_arr = np.append(vals_arr, np.array([['a/AU', tidx, fidx, eidx, bi, a_AU, a_AU_err]]), axis=0)

        # Make the DataFrame - cut off the first (all zeros) entries
        return pd.DataFrame(vals_arr[1:], columns=columns)

    def _print_results(self, result):
        '''
        Prints a results dict to terminal
        '''
        df = self._results_dict_to_dataframe(result, False)

        print(df)

    ###########################################################################
    #               PLOTTING THINGS                                           #
    ###########################################################################
    def _plot_samples(self, result, prior, fname, folder='./plots'):
        '''
        Makes a corner plot of the samples from a result
        '''
        samples = result.samples
        best = result.best
        weights = result.weights
        ndim = len(best)
        #labels = prior.fitting_params[:,0]
        labels= prior.get_latex_friendly_labels()

        titles=[] 
        for i in range(ndim):
            _l,_u = get_quantiles_on_best_val(samples[:,i], weights, best[i])
            _title = r'param = best$_{_l}^{_u}$'
            _title = _title.replace('param', labels[i])
            _title = _title.replace('best', f"{result.best[i]:.6f}")
            _title = _title.replace('_l', f"{_l:.6f}")
            _title = _title.replace('_u', f"{_u:.6f}")
            titles+=[_title]

        fig = corner.corner(samples, labels=labels, titles=titles,
                       show_titles=True, title_fmt=None, title_kwargs={"fontsize": 12})
        corner.overplot_lines(fig, best, color='green')

        # Add in the best value plots
        # Extract the axes
        #axes = np.array(fig.axes).reshape((ndim, ndim))
        # Loop over the diagonal
        #for i in range(ndim):
        #    ax = axes[i, i]
        #    ax.axvline(best[i], color="g")

        # Loop over the histograms
        """for yi in range(ndim):
            for xi in range(yi):
                #ax = axes[yi, xi]
                #ax.axvline(best[xi], color="g")
                #ax.axhline(best[yi], color="g")
                #ax.plot(best[xi], best[yi], "sg")"""

        fig.tight_layout()

        path = os.path.join(folder, fname)
        os.makedirs(os.path.dirname(path), exist_ok=True)

        fig.savefig(os.path.join(folder, fname), bbox_inches='tight', dpi=100)
        plt.close()

    def _plot_data(self, phase, flux, flux_err, model_phase, model_curve,
                       residuals, fname, title=None, folder='./plots',
                   figsize=(12,8), marker_color='dimgrey', line_color='black',phase_lims=None,
                       bin_data=False, cadence=2, binned_colour='red'):
            '''
            Plots the lightcurve and model consistently

            Parameters
            ----------

            cadence : float
                The cadence to bin to in phase, *not* minutes
            '''
            # Sort into phase orders
            data_order = np.argsort(phase)
            model_order = np.argsort(model_phase)

            phase = phase[data_order]
            flux = flux[data_order]
            if flux_err is not None:
                flux_err = flux_err[data_order]
            residuals = residuals[data_order]
            model_phase = model_phase[model_order]
            model_curve = model_curve[model_order]

            residuals_std = np.std(residuals)

            if bin_data:
                # Now we have to work out the binned values
                if flux_err is None:
                    err_for_bin = np.ones(len(flux))
                else:
                    err_for_bin = flux_err
                final_filter_lc = LightCurve(phase, flux, err_for_bin)
                binned_phase, binned_flux, binned_err, binned_residuals = final_filter_lc.bin(cadence, residuals)

                binned_residuals_std = np.std(binned_residuals)

                if flux_err is None:
                    binned_err = None

            # Set up the figure and the relevant axes
            gs = gridspec.GridSpec(6, 7)
            fig = plt.figure(figsize=figsize)

            main_ax = fig.add_subplot(gs[:-2, :-1])
            residual_ax = fig.add_subplot(gs[-2:, :-1], sharex=main_ax)
            hist_ax = fig.add_subplot(gs[-2:,-1], sharey=residual_ax)

            ####### PLOT! #########
            # Main plot
            main_ax.errorbar(phase, flux, flux_err,
                             zorder=1, capsize=2,
                             linestyle='', marker='.', color=marker_color,
                             elinewidth=0.8, alpha=0.6)

            main_ax.plot(model_phase, model_curve, zorder=3,
                         linewidth=2, color=line_color)
            if bin_data:
                main_ax.errorbar(binned_phase, binned_flux, binned_err,
                                 zorder=2, capsize=2, markersize=4,
                                 linestyle='', marker='.', color=binned_colour,
                                 elinewidth=0.8, alpha=0.6)
            if phase_lims!=None:
                main_ax.set_xlim(phase_lims)
            # Residuals
            residual_ax.errorbar(phase, residuals, flux_err, zorder=1,
                                 linestyle='', marker='.', capsize=2,
                                 color=marker_color, elinewidth=0.8,
                                 alpha=0.6)

            residual_ax.axhline(0, linestyle='dashed', color='gray',
                                linewidth=1, zorder=3)

            if bin_data:
                residual_ax.errorbar(binned_phase, binned_residuals,
                                     binned_err, zorder=2, markersize=4,
                                     linestyle='', marker='.', capsize=2,
                                     color=binned_colour, elinewidth=0.8,
                                     alpha=0.6)

            # Histogram residuals
            rgba_color = colors.to_rgba(marker_color)
            facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)

            unbinned_counts, bins = np.histogram(residuals, bins=30)

            hist_ax.hist(bins[:-1], bins, weights=unbinned_counts,
                         orientation='horizontal', color=facecolor,
                         edgecolor=rgba_color, histtype='stepfilled')
            hist_ax.axhline(0, linestyle='dashed', color='gray',
                            linewidth=1, zorder=1)


            if bin_data:
                rgba_color = colors.to_rgba(binned_colour)
                facecolor = (rgba_color[0], rgba_color[1], rgba_color[2], 0.6)

                binned_counts, _ = np.histogram(binned_residuals, bins)

                weighted_binned_counts = binned_counts * unbinned_counts.max()/binned_counts.max()

                hist_ax.hist(bins[:-1], bins, weights=weighted_binned_counts,
                             orientation='horizontal', color=facecolor,
                             edgecolor=rgba_color, histtype='stepfilled',alpha=0.5)

                hist_ax.text(0.02, 0.93, r'$\sigma_{unbinned} = $' + str(round(residuals_std, 5)), transform=hist_ax.transAxes)
                hist_ax.text(0.02, 0.85, r'$\sigma_{binned}$ = ' + str(round(binned_residuals_std, 5)), transform=hist_ax.transAxes)


            # FORMATTING AXES
            # Prune axes
            main_ax.yaxis.set_major_locator(MaxNLocator(6, prune='lower'))
            residual_ax.yaxis.set_major_locator(MaxNLocator(4, prune='upper'))
            #residual_ax.xaxis.set_major_locator(MaxNLocator(8, prune='upper'))

            # Add labels
            main_ax.set_ylabel('Normalised flux')
            residual_ax.set_ylabel('Residual')
            residual_ax.set_xlabel('Phase')

            # Format the axes
            main_ax.tick_params('both', which='both', direction='in',
                                labelbottom=False, top=True, right=True)

            residual_ax.tick_params('both', which='both', direction='in',
                                    top=True, right=True)

            hist_ax.tick_params('both', which='both', direction='in',
                                 labelleft=False, labelbottom=False,
                                 right=True, top=True)

            if title is not None:
                main_ax.set_title(title)

            fig.tight_layout()
            fig.subplots_adjust(hspace=0, wspace=0)

            os.makedirs(os.path.dirname(os.path.join(folder, fname)), exist_ok=True)

            # SAVE THE PLOT
            fig.savefig(os.path.join(folder, fname), bbox_inches='tight', dpi=300)

            plt.close()


class Results:
    """
    Dynesty>v 1.1 doesn't allow setting attributes directly.
    This new class help in creating a new instance with attributes for easier handling.
    """

    def __init__(self, sampler,):
        """sampler.results are the results from the dynesty sampler"""
        results = sampler.results

        self.logl = results.logl
        self.samples = results.samples
        self.logwt = results.logwt

        # Normalise weights
        self.weights = get_normalised_weights(results)

        # Calculate covariance matrix and use to get uncertainties
        cov = get_covariance_matrix(self)
        diagonal = np.diag(cov)
        uncertainties = np.sqrt(diagonal)

        self.cov = cov
        self.uncertainties = uncertainties

        # Get the 16th and 84th percentiles to use as upper and lower errors
        # This is arguably better than using the covariances(???)
        median = np.median(results.samples, axis=0)
        per_16 = np.percentile(results.samples, 16, axis=0)
        per_84 = np.percentile(results.samples, 84, axis=0)

        self.median = median
        self.lower_err = abs(median - per_16)
        self.upper_err = abs(per_84 - median)
        self.per_16 = per_16
        self.per_84 = per_84

        # Save the best fit results for easy access
        self.best = results.samples[np.argmax(results.logl)]


class ResultsException:
    """
    Handles results (same as above) during exception raised by dynesty.
    """

    def __init__(self, sampler,):
        """sampler.results are the results from the dynesty sampler"""
        results = sampler.results
        self.best = results.samples[np.argmax(results.logl)]



'''
TransitFit package

This package is designed to fit transit light curves using BATMAN
'''
name = 'transitfit'
__version__ = '3.4.1'

from .retriever import Retriever
from .priorinfo import PriorInfo
from .lightcurve import LightCurve
from ._pipeline import run_retrieval
from ._utils import split_lightcurve_file, calculate_logg, AU_to_host_radii
from .lightcurve import LightCurve
from .error_analysis import get_quantiles_on_best_val
